from __future__ import annotations

from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING

import pytest

from streamlink.stream.hls import (
    ByteRange,
    DateRange,
    ExtInf,
    HLSSegment,
    M3U8Parser,
    Media,
    Resolution,
    StreamInfo,
    parse_m3u8,
    parse_tag,
)
from tests.resources import text


if TYPE_CHECKING:
    from streamlink.stream.hls import M3U8, HLSPlaylist


UTC = timezone.utc


def test_parse_tag_mapping():
    class M3U8ParserSubclass(M3U8Parser):
        @parse_tag("EXT-X-VERSION")
        def parse_tag_ext_x_version(self):  # pragma: no cover
            pass

        @parse_tag("FOO-BAR")
        def parse_tag_foo_bar(self):  # pragma: no cover
            pass

    parent = M3U8Parser()
    assert hasattr(parent, "_TAGS")
    assert parent._TAGS["EXTINF"] is M3U8Parser.parse_tag_extinf
    assert parent._TAGS["EXT-X-VERSION"] is M3U8Parser.parse_tag_ext_x_version
    assert "FOO-BAR" not in parent._TAGS

    childA = M3U8ParserSubclass()
    assert hasattr(childA, "_TAGS")
    assert childA._TAGS["EXTINF"] is M3U8Parser.parse_tag_extinf
    assert childA._TAGS["EXT-X-VERSION"] is M3U8ParserSubclass.parse_tag_ext_x_version
    assert childA._TAGS["FOO-BAR"] is M3U8ParserSubclass.parse_tag_foo_bar

    assert parent._TAGS is not childA._TAGS

    childB = M3U8ParserSubclass()
    assert hasattr(childB, "_TAGS")
    assert childA._TAGS is childB._TAGS

    assert parent._TAGS["EXT-X-VERSION"].__doc__ is not None
    assert childA._TAGS["EXT-X-VERSION"].__doc__ is None


@pytest.mark.parametrize(
    ("string", "expected"),
    [
        ("", (None, None)),
        ("invalid", (None, None)),
        ("#TAG", ("TAG", "")),
        ("#TAG:ATTRIBUTES", ("TAG", "ATTRIBUTES")),
        ("#TAG:    ATTRIBUTES    ", ("TAG", "ATTRIBUTES")),
    ],
)
def test_split_tag(string: str, expected: tuple[str, str] | tuple[None, None]):
    assert M3U8Parser.split_tag(string) == expected


@pytest.mark.parametrize(
    ("attributes", "log", "expected"),
    [
        pytest.param("", False, {}, id="empty attribute list"),
        pytest.param("invalid", True, {}, id="invalid attribute list"),
        pytest.param("?=VALUE", True, {}, id="invalid attribute name"),
        pytest.param("key=VALUE", True, {}, id="lowercase attribute name"),
        pytest.param("KEY = VALUE", True, {}, id="invalid attribute format"),
        pytest.param("KEY=", True, {}, id="missing attribute value"),
        pytest.param('KEY="', True, {}, id="invalid attribute value"),
        pytest.param("KEY=123", False, {"KEY": "123"}, id="decimal integer"),
        pytest.param("KEY=0xdeadbeef", False, {"KEY": "0xdeadbeef"}, id="hexadecimal sequence, lowercase"),
        pytest.param("KEY=0XDEADBEEF", False, {"KEY": "0XDEADBEEF"}, id="hexadecimal sequence, uppercase"),
        pytest.param("KEY=123.456", False, {"KEY": "123.456"}, id="decimal floating point number"),
        pytest.param("KEY=-123.456", False, {"KEY": "-123.456"}, id="signed decimal floating point number"),
        pytest.param('KEY=""', False, {"KEY": ""}, id="empty quoted string"),
        pytest.param('KEY="foobar"', False, {"KEY": "foobar"}, id="quoted string"),
        pytest.param('KEY="foo"bar"', True, {}, id="invalid quoted string (quote)"),
        pytest.param('KEY="foo\rbar"', True, {}, id="invalid quoted string (carriage return)"),
        pytest.param('KEY="foo\nbar"', True, {}, id="invalid quoted string (new line)"),
        pytest.param("KEY=VALUE", False, {"KEY": "VALUE"}, id="enumerated string"),
        pytest.param("KEY=<@_@>", False, {"KEY": "<@_@>"}, id="enumerated string with special characters"),
        pytest.param("KEY=1920x1080", False, {"KEY": "1920x1080"}, id="decimal resolution"),
        pytest.param("KEY-123=VALUE", False, {"KEY-123": "VALUE"}, id="attribute name with alphanumerical chars and dashes"),
        pytest.param(
            'A=123,B=0xdeadbeef,C=123.456,D=-123.456,E="value, value, value",F=VALUE,G=1920x1080',
            False,
            {
                "A": "123",
                "B": "0xdeadbeef",
                "C": "123.456",
                "D": "-123.456",
                "E": "value, value, value",
                "F": "VALUE",
                "G": "1920x1080",
            },
            id="multiple attributes",
        ),
        pytest.param(
            'A="foo", B=123 , C=VALUE,D=456 ',
            False,
            {"A": "foo", "B": "123", "C": "VALUE", "D": "456"},
            id="multiple attributes with surrounding spaces (off-spec)",
        ),
        pytest.param(
            'A="foo"B=123',
            True,
            {},
            id="multiple attributes with invalid format (missing commas)",
        ),
        pytest.param(
            'A="foo",B = 123',
            True,
            {},
            id="multiple attributes with invalid format (spaces)",
        ),
        pytest.param(
            'A="foo",B=',
            True,
            {},
            id="multiple attributes with invalid format (missing value)",
        ),
    ],
)
def test_parse_attributes(caplog: pytest.LogCaptureFixture, attributes: str, log: bool, expected: dict):
    assert M3U8Parser.parse_attributes(attributes) == expected
    assert [(record.name, record.levelname, record.message) for record in caplog.records] == (
        [
            ("streamlink.stream.hls.m3u8", "warning", "Discarded invalid attributes list"),
        ]
        if log
        else []
    )


@pytest.mark.parametrize(
    ("string", "expected"),
    [
        ("", False),
        ("NO", False),
        ("YES", True),
    ],
)
def test_parse_bool(string: str, expected: bool):
    assert M3U8Parser.parse_bool(string) is expected


@pytest.mark.parametrize(
    ("string", "expected"),
    [
        ("", None),
        ("invalid", None),
        ("1234", ByteRange(1234, None)),
        ("1234@5678", ByteRange(1234, 5678)),
    ],
)
def test_parse_byterange(string: str, expected: ByteRange | None):
    assert M3U8Parser.parse_byterange(string) == expected


@pytest.mark.parametrize(
    ("string", "expected"),
    [
        ("", ExtInf(0, None)),
        ("invalid", ExtInf(0, None)),
        ("123", ExtInf(123.0, None)),
        ("123.456", ExtInf(123.456, None)),
        ("123.456,foo", ExtInf(123.456, "foo")),
    ],
)
def test_parse_extinf(string: str, expected: ExtInf):
    assert M3U8Parser.parse_extinf(string) == expected


@pytest.mark.parametrize(
    ("string", "signed", "expected", "log"),
    [
        (None, False, None, []),
        ("123", False, pytest.approx(123.0), []),
        ("123.", False, pytest.approx(123.0), []),
        ("123.456", False, pytest.approx(123.456), []),
        ("-123.456", False, pytest.approx(123.456), []),
        ("-123.456", True, pytest.approx(-123.456), []),
        (
            "",
            False,
            None,
            [("streamlink.stream.hls.m3u8", "warning", "Discarded invalid decimal-floating-point value")],
        ),
        (
            "foo",
            False,
            None,
            [("streamlink.stream.hls.m3u8", "warning", "Discarded invalid decimal-floating-point value")],
        ),
        (
            "123foo",
            False,
            None,
            [("streamlink.stream.hls.m3u8", "warning", "Discarded invalid decimal-floating-point value")],
        ),
        (
            "",
            True,
            None,
            [("streamlink.stream.hls.m3u8", "warning", "Discarded invalid signed-decimal-floating-point value")],
        ),
        (
            "foo",
            True,
            None,
            [("streamlink.stream.hls.m3u8", "warning", "Discarded invalid signed-decimal-floating-point value")],
        ),
        (
            "123foo",
            True,
            None,
            [("streamlink.stream.hls.m3u8", "warning", "Discarded invalid signed-decimal-floating-point value")],
        ),
    ],
)
def test_parse_float(caplog: pytest.LogCaptureFixture, string: str | None, signed: bool, expected: bytes | None, log: list):
    assert M3U8Parser.parse_float(string, signed=signed) == expected
    assert [(record.name, record.levelname, record.message) for record in caplog.records] == log


@pytest.mark.parametrize(
    ("string", "log", "expected"),
    [
        (None, False, None),
        ("", True, None),
        ("deadbeef", True, None),
        ("0xnothex", True, None),
        ("0xdeadbeef", False, b"\xde\xad\xbe\xef"),
        ("0XDEADBEEF", False, b"\xde\xad\xbe\xef"),
        ("0xdeadbee", False, b"\x0d\xea\xdb\xee"),
    ],
)
def test_parse_hex(caplog: pytest.LogCaptureFixture, string: str | None, log: bool, expected: bytes | None):
    assert M3U8Parser.parse_hex(string) == expected
    assert [(record.name, record.levelname, record.message) for record in caplog.records] == (
        [
            ("streamlink.stream.hls.m3u8", "warning", "Discarded invalid hexadecimal-sequence attribute value"),
        ]
        if log
        else []
    )


@pytest.mark.parametrize(
    ("string", "log", "expected"),
    [
        (None, False, None),
        ("not an ISO8601 string", True, None),
        ("2000-01-01", True, None),
        ("2000-99-99T99:99:99.999Z", True, None),
        ("2000-01-01T00:00:00.000Z", False, datetime(2000, 1, 1, 0, 0, 0, 0, tzinfo=UTC)),
    ],
)
def test_parse_iso8601(caplog: pytest.LogCaptureFixture, string: str | None, log: bool, expected: datetime | None):
    assert M3U8Parser.parse_iso8601(string) == expected
    assert [(record.name, record.levelname, record.message) for record in caplog.records] == (
        [
            ("streamlink.stream.hls.m3u8", "warning", "Discarded invalid ISO8601 attribute value"),
        ]
        if log
        else []
    )


@pytest.mark.parametrize(
    ("string", "expected"),
    [
        (None, None),
        ("123", timedelta(seconds=123.0)),
        ("123.456", timedelta(seconds=123.456)),
        ("-123.456", timedelta(seconds=-123.456)),
    ],
)
def test_parse_timedelta(string: str | None, expected: timedelta | None):
    assert M3U8Parser.parse_timedelta(string) == expected


@pytest.mark.parametrize(
    ("string", "expected"),
    [
        ("", Resolution(0, 0)),
        ("invalid", Resolution(0, 0)),
        ("1920x1080", Resolution(1920, 1080)),
    ],
)
def test_parse_resolution(string: str, expected: Resolution):
    assert M3U8Parser.parse_resolution(string) == expected


class TestHLSPlaylist:
    def test_load(self) -> None:
        with text("hls/test_1.m3u8") as m3u8_fh:
            playlist = parse_m3u8(m3u8_fh.read(), "http://test.se/", parser=M3U8Parser)

        assert playlist.media == [
            Media(
                uri="http://test.se/audio/stereo/en/128kbit.m3u8",
                type="AUDIO",
                group_id="stereo",
                language="en",
                name="English",
                default=True,
                autoselect=True,
                forced=False,
                characteristics=None,
            ),
            Media(
                uri="http://test.se/audio/stereo/none/128kbit.m3u8",
                type="AUDIO",
                group_id="stereo",
                language="zxx",
                name="Dubbing",
                default=False,
                autoselect=True,
                forced=False,
                characteristics=None,
            ),
            Media(
                uri="http://test.se/audio/surround/en/320kbit.m3u8",
                type="AUDIO",
                group_id="surround",
                language="en",
                name="English",
                default=True,
                autoselect=True,
                forced=False,
                characteristics=None,
            ),
            Media(
                uri="http://test.se/audio/stereo/none/128kbit.m3u8",
                type="AUDIO",
                group_id="surround",
                language="zxx",
                name="Dubbing",
                default=False,
                autoselect=True,
                forced=False,
                characteristics=None,
            ),
            Media(
                uri="http://test.se/subtitles_de.m3u8",
                type="SUBTITLES",
                group_id="subs",
                language="de",
                name="Deutsch",
                default=False,
                autoselect=True,
                forced=False,
                characteristics=None,
            ),
            Media(
                uri="http://test.se/subtitles_en.m3u8",
                type="SUBTITLES",
                group_id="subs",
                language="en",
                name="English",
                default=True,
                autoselect=True,
                forced=False,
                characteristics=None,
            ),
            Media(
                uri="http://test.se/subtitles_es.m3u8",
                type="SUBTITLES",
                group_id="subs",
                language="es",
                name="Espanol",
                default=False,
                autoselect=True,
                forced=False,
                characteristics=None,
            ),
            Media(
                uri="http://test.se/subtitles_fr.m3u8",
                type="SUBTITLES",
                group_id="subs",
                language="fr",
                name="Français",
                default=False,
                autoselect=True,
                forced=False,
                characteristics=None,
            ),
        ]

        assert [p.stream_info for p in playlist.playlists] == [
            StreamInfo(
                bandwidth=260000,
                program_id="1",
                codecs=["avc1.4d400d", "mp4a.40.2"],
                resolution=Resolution(width=422, height=180),
                framerate=None,
                audio="stereo",
                video=None,
                subtitles="subs",
            ),
            StreamInfo(
                bandwidth=520000,
                program_id="1",
                codecs=["avc1.4d4015", "mp4a.40.2"],
                resolution=Resolution(width=638, height=272),
                framerate=None,
                audio="stereo",
                video=None,
                subtitles="subs",
            ),
            StreamInfo(
                bandwidth=830000,
                program_id="1",
                codecs=["avc1.4d4015", "mp4a.40.2"],
                resolution=Resolution(width=638, height=272),
                framerate=None,
                audio="stereo",
                video=None,
                subtitles="subs",
            ),
            StreamInfo(
                bandwidth=1100000,
                program_id="1",
                codecs=["avc1.4d401f", "mp4a.40.2"],
                resolution=Resolution(width=958, height=408),
                framerate=None,
                audio="surround",
                video=None,
                subtitles="subs",
            ),
            StreamInfo(
                bandwidth=1600000,
                program_id="1",
                codecs=["avc1.4d401f", "mp4a.40.2"],
                resolution=Resolution(width=1277, height=554),
                framerate=None,
                audio="surround",
                video=None,
                subtitles="subs",
            ),
            StreamInfo(
                bandwidth=4100000,
                program_id="1",
                codecs=["avc1.4d4028", "mp4a.40.2"],
                resolution=Resolution(width=1921, height=818),
                framerate=None,
                audio="surround",
                video=None,
                subtitles="subs",
            ),
            StreamInfo(
                bandwidth=6200000,
                program_id="1",
                codecs=["avc1.4d4028", "mp4a.40.2"],
                resolution=Resolution(width=1921, height=818),
                framerate=None,
                audio="surround",
                video=None,
                subtitles="subs",
            ),
            StreamInfo(
                bandwidth=10000000,
                program_id="1",
                codecs=["avc1.4d4033", "mp4a.40.2"],
                resolution=Resolution(width=4096, height=1744),
                framerate=None,
                audio="surround",
                video=None,
                subtitles="subs",
            ),
        ]

    def test_parse_date(self) -> None:
        with text("hls/test_date.m3u8") as m3u8_fh:
            playlist = parse_m3u8(m3u8_fh.read(), "http://test.se/", parser=M3U8Parser)

        start_date = datetime(year=2000, month=1, day=1, hour=0, minute=0, second=0, microsecond=0, tzinfo=UTC)
        end_date = datetime(year=2000, month=1, day=1, hour=0, minute=1, second=0, microsecond=0, tzinfo=UTC)
        delta_15 = timedelta(seconds=15)
        delta_30 = timedelta(seconds=30, milliseconds=500)
        delta_60 = timedelta(seconds=60)

        assert playlist.targetduration == 120

        assert list(playlist.dateranges) == [
            DateRange(
                id="start-invalid",
                start_date=None,
                classname=None,
                end_date=None,
                duration=None,
                planned_duration=None,
                end_on_next=False,
                x={},
            ),
            DateRange(
                id="start-no-frac",
                start_date=start_date,
                classname=None,
                end_date=None,
                duration=None,
                planned_duration=None,
                end_on_next=False,
                x={},
            ),
            DateRange(
                id="start-with-frac",
                start_date=start_date,
                classname=None,
                end_date=None,
                duration=None,
                planned_duration=None,
                end_on_next=False,
                x={},
            ),
            DateRange(
                id="with-class",
                start_date=start_date,
                classname="bar",
                end_date=None,
                duration=None,
                planned_duration=None,
                end_on_next=False,
                x={},
            ),
            DateRange(
                id="duration",
                start_date=start_date,
                duration=delta_30,
                classname=None,
                end_date=None,
                planned_duration=None,
                end_on_next=False,
                x={},
            ),
            DateRange(
                id="planned-duration",
                start_date=start_date,
                planned_duration=delta_15,
                classname=None,
                end_date=None,
                duration=None,
                end_on_next=False,
                x={},
            ),
            DateRange(
                id="duration-precedence",
                start_date=start_date,
                duration=delta_30,
                planned_duration=delta_15,
                classname=None,
                end_date=None,
                end_on_next=False,
                x={},
            ),
            DateRange(
                id="end",
                start_date=start_date,
                end_date=end_date,
                classname=None,
                duration=None,
                planned_duration=None,
                end_on_next=False,
                x={},
            ),
            DateRange(
                id="end-precedence",
                start_date=start_date,
                end_date=end_date,
                duration=delta_30,
                classname=None,
                planned_duration=None,
                end_on_next=False,
                x={},
            ),
            DateRange(
                id=None,
                start_date=None,
                end_date=None,
                duration=None,
                classname=None,
                planned_duration=None,
                end_on_next=False,
                x={"X-CUSTOM": "value"},
            ),
        ]
        assert list(playlist.segments) == [
            HLSSegment(
                uri="http://test.se/segment0-15.ts",
                num=0,
                duration=15.0,
                title="live",
                date=start_date,
                key=None,
                discontinuity=False,
                byterange=None,
                map=None,
            ),
            HLSSegment(
                uri="http://test.se/segment15-30.5.ts",
                num=1,
                duration=15.5,
                title="live",
                date=start_date + delta_15,
                key=None,
                discontinuity=False,
                byterange=None,
                map=None,
            ),
            HLSSegment(
                uri="http://test.se/segment30.5-60.ts",
                num=2,
                duration=29.5,
                title="live",
                date=start_date + delta_30,
                key=None,
                discontinuity=False,
                byterange=None,
                map=None,
            ),
            HLSSegment(
                uri="http://test.se/segment60-.ts",
                num=3,
                duration=60.0,
                title="live",
                date=start_date + delta_60,
                key=None,
                discontinuity=False,
                byterange=None,
                map=None,
            ),
        ]

        assert [playlist.is_date_in_daterange(playlist.segments[0].date, daterange) for daterange in playlist.dateranges] \
               == [None, True, True, True, True, True, True, True, True, None]  # fmt: skip
        assert [playlist.is_date_in_daterange(playlist.segments[1].date, daterange) for daterange in playlist.dateranges] \
               == [None, True, True, True, True, False, True, True, True, None]  # fmt: skip
        assert [playlist.is_date_in_daterange(playlist.segments[2].date, daterange) for daterange in playlist.dateranges] \
               == [None, True, True, True, False, False, False, True, True, None]  # fmt: skip
        assert [playlist.is_date_in_daterange(playlist.segments[3].date, daterange) for daterange in playlist.dateranges] \
               == [None, True, True, True, False, False, False, False, False, None]  # fmt: skip

    def test_parse_bandwidth(self) -> None:
        with text("hls/test_multivariant_bandwidth.m3u8") as m3u8_fh:
            playlist: M3U8[HLSSegment, HLSPlaylist] = parse_m3u8(m3u8_fh.read(), "http://mocked/", parser=M3U8Parser)

        assert [(pl.stream_info.video, pl.stream_info.bandwidth) for pl in playlist.playlists] == [
            ("chunked", 0),
            ("audio_only", 0),
            ("720p60", 3000000),
            ("480p30", 1500000),
            ("360p30", 700000),
            ("160p30", 300000),
        ]



from __future__ import annotations

import socket
import ssl
from operator import itemgetter
from socket import AF_INET, AF_INET6
from ssl import SSLContext
from typing import TYPE_CHECKING
from unittest.mock import Mock, PropertyMock, call

import freezegun
import pytest
import requests
import requests_mock as rm
import urllib3
from requests.adapters import HTTPAdapter
from urllib3.connection import HTTPConnection
from urllib3.response import HTTPResponse

from streamlink.exceptions import PluginError, StreamlinkDeprecationWarning
from streamlink.session.http import (
    HTTPSession,
    SSLContextAdapter,
    TLSNoDHAdapter,
    TLSSecLevel1Adapter,
    urllib3_set_socket_options,
)
from streamlink.session.http_useragents import DEFAULT


if TYPE_CHECKING:
    from pathlib import Path

    from streamlink import Streamlink


_original_allowed_gai_family = urllib3.util.connection.allowed_gai_family  # type: ignore[attr-defined, ty:unresolved-attribute]


class TestUrllib3Overrides:
    @pytest.fixture(scope="class")
    def httpsession(self) -> HTTPSession:
        return HTTPSession()

    @pytest.mark.parametrize(
        ("url", "expected"),
        [
            pytest.param(
                "https://foo/bar%3F?baz%21",
                "https://foo/bar%3F?baz%21",
                id="keep-encoded-reserved-characters",
            ),
            pytest.param(
                "https://foo/%62%61%72?%62%61%7A",
                "https://foo/bar?baz",
                id="decode-encoded-unreserved-characters",
            ),
            pytest.param(
                "https://foo/bär?bäz",
                "https://foo/b%C3%A4r?b%C3%A4z",
                id="encode-other-characters",
            ),
            pytest.param(
                "https://foo/b%c3%a4r?b%c3%a4z",
                "https://foo/b%c3%a4r?b%c3%a4z",
                id="keep-percent-encodings-with-lowercase-characters",
            ),
            pytest.param(
                "https://foo/b%C3%A4r?b%C3%A4z",
                "https://foo/b%C3%A4r?b%C3%A4z",
                id="keep-percent-encodings-with-uppercase-characters",
            ),
            pytest.param(
                "https://foo/%?%",
                "https://foo/%25?%25",
                id="empty-percent-encodings-without-valid-encodings",
            ),
            pytest.param(
                "https://foo/%0?%0",
                "https://foo/%250?%250",
                id="incomplete-percent-encodings-without-valid-encodings",
            ),
            pytest.param(
                "https://foo/%zz?%zz",
                "https://foo/%25zz?%25zz",
                id="invalid-percent-encodings-without-valid-encodings",
            ),
            pytest.param(
                "https://foo/%3F%?%3F%",
                "https://foo/%253F%25?%253F%25",
                id="empty-percent-encodings-with-valid-encodings",
            ),
            pytest.param(
                "https://foo/%3F%0?%3F%0",
                "https://foo/%253F%250?%253F%250",
                id="incomplete-percent-encodings-with-valid-encodings",
            ),
            pytest.param(
                "https://foo/%3F%zz?%3F%zz",
                "https://foo/%253F%25zz?%253F%25zz",
                id="invalid-percent-encodings-with-valid-encodings",
            ),
        ],
    )
    def test_encode_invalid_chars(self, httpsession: HTTPSession, url: str, expected: str):
        req = requests.Request(method="GET", url=url)
        prep = httpsession.prepare_request(req)
        assert prep.url == expected

    @pytest.mark.parametrize(
        ("sock", "options", "expected"),
        [
            pytest.param(
                {},
                [],
                [],
                id="no-options",
            ),
            pytest.param(
                {},
                [(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)],
                [(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)],
                id="no-filters",
            ),
            # see set_interface() on darwin
            pytest.param(
                {},
                [(socket.IPPROTO_IP, 25, 1), (socket.IPPROTO_IPV6, 125, 1)],
                [(socket.IPPROTO_IP, 25, 1)],
                id="inet-filter-ipproto-ipv6",
            ),
            # see set_interface() on darwin
            pytest.param(
                {"family": socket.AF_INET6},
                [(socket.IPPROTO_IP, 25, 1), (socket.IPPROTO_IPV6, 125, 1)],
                [(socket.IPPROTO_IPV6, 125, 1)],
                id="inet6-filter-ipproto-ip",
            ),
        ],
    )
    def test_set_socket_options(self, sock: dict, options: list, expected: list):
        sock.setdefault("family", socket.AF_INET)
        sock.setdefault("type", socket.SOCK_STREAM)
        sock.setdefault("proto", socket.IPPROTO_TCP)
        mock_socket = Mock(**sock)
        urllib3_set_socket_options(mock_socket, options)
        assert mock_socket.setsockopt.call_args_list == [call(*item) for item in expected]


class TestHTTPSession:
    def test_session_init(self):
        session = HTTPSession()
        assert session.headers.get("User-Agent") == DEFAULT
        assert session.timeout == 20.0  # noqa: RUF069
        assert "file://" in session.adapters.keys()

    def test_read_timeout(self, monkeypatch: pytest.MonkeyPatch):
        mock_sleep = Mock()
        mock_request = Mock(side_effect=requests.Timeout)
        monkeypatch.setattr("streamlink.session.http.time.sleep", mock_sleep)
        monkeypatch.setattr("streamlink.session.http.Session.request", mock_request)

        session = HTTPSession()
        with pytest.raises(PluginError, match=r"^Unable to open URL: http://localhost/"):
            session.get("http://localhost/", timeout=123, retries=3, retry_backoff=2, retry_max_backoff=5)

        assert mock_request.call_args_list == [
            call("GET", "http://localhost/", headers={}, params={}, timeout=123, proxies={}, allow_redirects=True),
            call("GET", "http://localhost/", headers={}, params={}, timeout=123, proxies={}, allow_redirects=True),
            call("GET", "http://localhost/", headers={}, params={}, timeout=123, proxies={}, allow_redirects=True),
            call("GET", "http://localhost/", headers={}, params={}, timeout=123, proxies={}, allow_redirects=True),
        ]
        assert mock_sleep.call_args_list == [
            call(2),
            call(4),
            call(5),
        ]

    @pytest.mark.parametrize("encoding", ["UTF-32BE", "UTF-32LE", "UTF-16BE", "UTF-16LE", "UTF-8"])
    def test_determine_json_encoding(self, recwarn: pytest.WarningsRecorder, encoding: str):
        data = "Hello world, Γειά σου Κόσμε, こんにちは世界".encode(encoding)  # noqa: RUF001
        assert HTTPSession.determine_json_encoding(data) == encoding
        assert [(record.category, str(record.message)) for record in recwarn.list] == [
            (StreamlinkDeprecationWarning, "Deprecated HTTPSession.determine_json_encoding() call"),
        ]

    @pytest.mark.parametrize(
        ("encoding", "override"),
        [
            ("utf-32-be", None),
            ("utf-32-le", None),
            ("utf-16-be", None),
            ("utf-16-le", None),
            ("utf-8", None),
            # With byte order mark (BOM)
            ("utf-16", None),
            ("utf-32", None),
            ("utf-8-sig", None),
            # Override
            ("utf-8", "utf-8"),
            ("cp949", "cp949"),
        ],
    )
    def test_json(self, monkeypatch: pytest.MonkeyPatch, encoding: str, override: str | None):
        mock_content = PropertyMock(return_value='{"test": "Α and Ω"}'.encode(encoding))  # noqa: RUF001
        monkeypatch.setattr("requests.Response.content", mock_content)

        res = requests.Response()
        res.encoding = override

        assert HTTPSession.json(res) == {"test": "Α and Ω"}  # noqa: RUF001

    @pytest.mark.parametrize(
        ("content_type", "encoding", "content", "expected"),
        [
            pytest.param(
                "text/html",
                None,
                b"B\xe4r",
                "ISO-8859-1",
                id="default-iso-8859-1-charset-for-text",
            ),
            pytest.param(
                "application/json",
                None,
                b"B\xc3\xa4r",
                "utf-8",
                id="default-utf-8-charset-for-json",
            ),
            pytest.param(
                'text/html; charset="ISO-8859-1"',
                None,
                b"B\xe4r",
                "ISO-8859-1",
                id="declared-iso-8859-1-charset",
            ),
            pytest.param(
                'text/html; charset="utf-8"',
                None,
                b"B\xc3\xa4r",
                "utf-8",
                id="declared-utf-8-charset",
            ),
            pytest.param(
                "text/html",
                "utf-8",
                b"B\xc3\xa4r",
                "utf-8",
                id="override-missing-charset",
            ),
            pytest.param(
                'text/html; charset="ISO-8859-1"',
                "utf-8",
                b"B\xc3\xa4r",
                "utf-8",
                id="override-incorrect-charset",
            ),
            pytest.param(
                'text/html; charset="utf-8"',
                "utf-8",
                b"B\xc3\xa4r",
                "utf-8",
                id="override-same-charset",
            ),
        ],
    )
    @pytest.mark.parametrize("method", ["get", "post", "head", "put", "patch", "delete"])
    def test_encoding_override(
        self,
        requests_mock: rm.Mocker,
        method: str,
        content_type: str,
        encoding: str,
        content: bytes,
        expected: str,
    ):
        requests_mock.register_uri(rm.ANY, "http://mocked", headers={"Content-Type": content_type}, content=content)
        httpsession = HTTPSession()
        res = getattr(httpsession, method)("http://mocked", encoding=encoding)
        assert res.encoding == expected
        assert res.text == "Bär"

    # Linux
    SOL_SOCKET = getattr(socket, "SOL_SOCKET", 1)
    SO_BINDTODEVICE = getattr(socket, "SO_BINDTODEVICE", 25)

    # Darwin
    IPPROTO_IP = getattr(socket, "IPPROTO_IP", 0)
    IPPROTO_IPV6 = getattr(socket, "IPPROTO_IPV6", 41)
    IP_BOUND_IF = getattr(socket, "IP_BOUND_IF", 25)
    IPV6_BOUND_IF = getattr(socket, "IPV6_BOUND_IF", 125)

    @pytest.mark.parametrize(
        ("interface", "source_address", "socket_options", "log"),
        [
            pytest.param(
                None,
                None,
                None,
                [],
                id="none",
            ),
            pytest.param(
                "",
                None,
                None,
                [],
                id="empty",
            ),
            pytest.param(
                "0.0.0.0",
                ("0.0.0.0", 0),
                None,
                [],
                id="ipv4",
            ),
            pytest.param(
                "::1",
                ("::1", 0),
                None,
                [],
                id="ipv6",
            ),
            pytest.param(
                "my-interface",
                None,
                [
                    *HTTPConnection.default_socket_options,
                    (SOL_SOCKET, SO_BINDTODEVICE, b"my-interface"),
                ],
                [],
                id="linux-iface",
                marks=pytest.mark.linux_only,
            ),
            pytest.param(
                "if!my-interface",
                None,
                [
                    *HTTPConnection.default_socket_options,
                    (SOL_SOCKET, SO_BINDTODEVICE, b"my-interface"),
                ],
                [],
                id="linux-iface-prefix",
                marks=pytest.mark.linux_only,
            ),
            pytest.param(
                "my-interface",
                None,
                [
                    *HTTPConnection.default_socket_options,
                    (IPPROTO_IP, IP_BOUND_IF, 1),
                    (IPPROTO_IPV6, IPV6_BOUND_IF, 1),
                ],
                [],
                id="darwin-iface",
                marks=pytest.mark.darwin_only,
            ),
            pytest.param(
                "if!my-interface",
                None,
                [
                    *HTTPConnection.default_socket_options,
                    (IPPROTO_IP, IP_BOUND_IF, 1),
                    (IPPROTO_IPV6, IPV6_BOUND_IF, 1),
                ],
                [],
                id="darwin-iface-prefix",
                marks=pytest.mark.darwin_only,
            ),
            pytest.param(
                "host!0.0.0.0",
                ("0.0.0.0", 0),
                None,
                [],
                id="unix-host-prefix",
                marks=pytest.mark.posix_only,
            ),
            pytest.param(
                "host!foo",
                ("foo", 0),
                None,
                [],
                id="unix-host-prefix-hostname",
                marks=pytest.mark.posix_only,
            ),
            pytest.param(
                "ifhost!my-interface!0.0.0.0",
                ("0.0.0.0", 0),
                [
                    *HTTPConnection.default_socket_options,
                    (SOL_SOCKET, SO_BINDTODEVICE, b"my-interface"),
                ],
                [],
                id="linux-ifhost-prefix",
                marks=pytest.mark.linux_only,
            ),
            pytest.param(
                "ifhost!my-interface",
                None,
                [
                    *HTTPConnection.default_socket_options,
                    (SOL_SOCKET, SO_BINDTODEVICE, b"ifhost!my-interface"),
                ],
                [],
                id="linux-ifhost-prefix-invalid",
                marks=pytest.mark.linux_only,
            ),
            pytest.param(
                "ifhost!my-interface!0.0.0.0",
                ("0.0.0.0", 0),
                [
                    *HTTPConnection.default_socket_options,
                    (IPPROTO_IP, IP_BOUND_IF, 1),
                    (IPPROTO_IPV6, IPV6_BOUND_IF, 1),
                ],
                [],
                id="darwin-ifhost-prefix",
                marks=pytest.mark.darwin_only,
            ),
            pytest.param(
                "ifhost!my-interface",
                None,
                None,
                [
                    ("streamlink.session.http", "error", "Invalid network interface name"),
                ],
                id="darwin-ifhost-prefix-invalid",
                marks=pytest.mark.darwin_only,
            ),
            pytest.param(
                "my-interface",
                ("my-interface", 0),
                None,
                [],
                id="win32-iface",
                marks=pytest.mark.windows_only,
            ),
            pytest.param(
                "ifhost!my-interface!0.0.0.0",
                ("ifhost!my-interface!0.0.0.0", 0),
                None,
                [],
                id="win32-prefix",
                marks=pytest.mark.windows_only,
            ),
        ],
    )
    def test_set_interface(
        self,
        monkeypatch: pytest.MonkeyPatch,
        caplog: pytest.LogCaptureFixture,
        interface: str,
        source_address: tuple | None,
        socket_options: list[tuple] | None,
        log: list,
    ):
        def _mock_socket_if_nametoindex(iface: str):
            match iface:
                case "my-interface":
                    return 1
                case _:
                    raise OSError("Invalid network interface name")

        monkeypatch.setattr("socket.if_nametoindex", _mock_socket_if_nametoindex)

        session = HTTPSession()
        session.mount("custom://", TLSNoDHAdapter())

        a_http, a_https, a_custom, a_file = itemgetter("http://", "https://", "custom://", "file://")(session.adapters)
        assert isinstance(a_http, HTTPAdapter)
        assert isinstance(a_https, HTTPAdapter)
        assert isinstance(a_custom, HTTPAdapter)
        assert not isinstance(a_file, HTTPAdapter)

        for adapter in a_http, a_https, a_custom:
            assert adapter.poolmanager.connection_pool_kw.get("source_address") is None
            assert adapter.poolmanager.connection_pool_kw.get("socket_options") is None

        session.set_interface(interface=interface)
        for adapter in a_http, a_https, a_custom:
            assert adapter.poolmanager.connection_pool_kw.get("source_address") == source_address
            assert adapter.poolmanager.connection_pool_kw.get("socket_options") == socket_options
        assert [(record.name, record.levelname, record.message) for record in caplog.records] == log

        a_new = TLSNoDHAdapter()
        assert a_new.poolmanager.connection_pool_kw.get("source_address") is None
        assert a_new.poolmanager.connection_pool_kw.get("socket_options") is None
        session.mount("new://", a_new)
        assert a_new.poolmanager.connection_pool_kw.get("source_address") == source_address
        assert a_new.poolmanager.connection_pool_kw.get("socket_options") == socket_options

        session.set_interface(interface="")
        for adapter in a_http, a_https, a_custom, a_new:
            assert adapter.poolmanager.connection_pool_kw.get("source_address") is None
            assert adapter.poolmanager.connection_pool_kw.get("socket_options") is None

        session.set_interface(interface=None)
        for adapter in a_http, a_https, a_custom, a_new:
            assert adapter.poolmanager.connection_pool_kw.get("source_address") is None
            assert adapter.poolmanager.connection_pool_kw.get("socket_options") is None

        # doesn't raise
        session.set_interface(interface=None)

    def test_set_address_family(self, monkeypatch: pytest.MonkeyPatch):
        session = HTTPSession()
        mock_urllib3_util_connection = Mock(allowed_gai_family=_original_allowed_gai_family)
        monkeypatch.setattr("streamlink.session.http.urllib3_util_connection", mock_urllib3_util_connection)

        assert mock_urllib3_util_connection.allowed_gai_family is _original_allowed_gai_family

        session.set_address_family(family=AF_INET)
        assert mock_urllib3_util_connection.allowed_gai_family is not _original_allowed_gai_family
        assert mock_urllib3_util_connection.allowed_gai_family() is AF_INET

        session.set_address_family(family=None)
        assert mock_urllib3_util_connection.allowed_gai_family is _original_allowed_gai_family

        session.set_address_family(family=AF_INET6)
        assert mock_urllib3_util_connection.allowed_gai_family is not _original_allowed_gai_family
        assert mock_urllib3_util_connection.allowed_gai_family() is AF_INET6

        session.set_address_family(family=None)
        assert mock_urllib3_util_connection.allowed_gai_family is _original_allowed_gai_family

    def test_disable_dh(self):
        session = HTTPSession()

        assert isinstance(session.adapters["https://"], HTTPAdapter)
        assert not isinstance(session.adapters["https://"], TLSNoDHAdapter)

        assert not session.adapters["https://"].poolmanager.connection_pool_kw.get("source_address")
        session.adapters["https://"].poolmanager.connection_pool_kw.update(source_address=("0.0.0.0", 0))

        session.disable_dh(disable=True)
        assert isinstance(session.adapters["https://"], TLSNoDHAdapter)
        assert session.adapters["https://"].poolmanager.connection_pool_kw.get("source_address") == ("0.0.0.0", 0)

        session.disable_dh(disable=False)
        assert isinstance(session.adapters["https://"], HTTPAdapter)
        assert not isinstance(session.adapters["https://"], TLSNoDHAdapter)
        assert session.adapters["https://"].poolmanager.connection_pool_kw.get("source_address") == ("0.0.0.0", 0)


class TestHTTPCookies:
    def test_invalid_file(self, tmp_path: Path):
        session = HTTPSession()
        cookies_file = tmp_path / "invalid-file"
        with pytest.raises(FileNotFoundError, match=r" is not a valid cookies file path$"):
            session.set_cookies_from_file(cookies_file)
        assert session.cookies == {}

    def test_invalid_format(self, tmp_path: Path):
        session = HTTPSession()
        cookies_file = tmp_path / "invalid-file"
        cookies_file.write_text("""foo\nbar\n""")
        with pytest.raises(Exception, match=r" does not look like a Netscape format cookies file$"):
            session.set_cookies_from_file(cookies_file)
        assert session.cookies == {}

    @freezegun.freeze_time("2000-01-01T00:00:00Z")
    def test_cookies(self, tmp_path: Path, requests_mock: rm.Mocker):
        session = HTTPSession()
        cookies_file = tmp_path / "cookies.txt"
        content = "".join([
            "# Netscape HTTP Cookie File\n",
            "host.local	FALSE	/a	FALSE	946684801	foo	bar\n",
            "host.local	FALSE	/b	FALSE	946684801	baz	qux\n",
            "host.local	FALSE	/	FALSE	946684799	expired	1\n",
            ".host.tld	TRUE	/	TRUE	946684801	abc	def\n",
        ])
        cookies_file.write_text(content)
        session.set_cookies_from_file(cookies_file)
        assert session.cookies.get_dict(domain="host.local") == {
            "foo": "bar",
            "baz": "qux",
        }
        assert session.cookies.get_dict(domain="host.local", path="/a") == {
            "foo": "bar",
        }
        assert session.cookies.get_dict(domain=".host.tld") == {
            "abc": "def",
        }

        matcher = requests_mock.register_uri(rm.ANY, rm.ANY, status_code=200)

        # domain, path, expired
        assert session.get("http://host.local/a").status_code == 200
        assert matcher.request_history[-1]._request.headers.get("cookie") == "foo=bar"
        assert session.get("http://host.local/b").status_code == 200
        assert matcher.request_history[-1]._request.headers.get("cookie") == "baz=qux"
        assert session.get("http://foo.host.local/").status_code == 200
        assert matcher.request_history[-1]._request.headers.get("cookie") is None

        # subdomain, secure
        assert session.get("https://sub.host.tld/").status_code == 200
        assert matcher.request_history[-1]._request.headers.get("cookie") == "abc=def"
        assert session.get("http://sub.host.tld/").status_code == 200
        assert matcher.request_history[-1]._request.headers.get("cookie") is None

    @freezegun.freeze_time("2000-01-01T00:00:00Z")
    def test_cookies_override(self, tmp_path: Path):
        session = HTTPSession()
        file_a = tmp_path / "a"
        file_b = tmp_path / "b"
        file_a.write_text(
            "".join([
                "# Netscape HTTP Cookie File\n",
                "host.local	FALSE	/	FALSE	946684801	foo	bar\n",
                "host.local	FALSE	/	FALSE	946684801	abc	def\n",
            ]),
        )
        file_b.write_text(
            "".join([
                "# Netscape HTTP Cookie File\n",
                "host.local	FALSE	/	FALSE	946684801	foo	baz\n",
            ]),
        )

        assert session.cookies.get_dict(domain="host.local") == {}
        session.set_cookies_from_file(file_a)
        assert session.cookies.get_dict(domain="host.local") == {
            "foo": "bar",
            "abc": "def",
        }
        session.set_cookies_from_file(file_b)
        assert session.cookies.get_dict(domain="host.local") == {
            "foo": "baz",
            "abc": "def",
        }


class TestHTTPAdapters:
    @staticmethod
    def _has_dh_ciphers(ssl_context: SSLContext):
        return any(cipher["kea"] == "kx-dhe" for cipher in ssl_context.get_ciphers())

    @staticmethod
    def _has_weak_digest_ciphers(ssl_context: SSLContext):
        return any(cipher["digest"] == "sha1" for cipher in ssl_context.get_ciphers())

    def test_sslcontextadapter(self):
        adapter = SSLContextAdapter()
        ssl_context = adapter.poolmanager.connection_pool_kw.get("ssl_context")
        assert isinstance(ssl_context, SSLContext)
        assert self._has_dh_ciphers(ssl_context)
        assert not self._has_weak_digest_ciphers(ssl_context)

    def test_tlsnodhadapter(self):
        adapter = TLSNoDHAdapter()
        ssl_context = adapter.poolmanager.connection_pool_kw.get("ssl_context")
        assert isinstance(ssl_context, SSLContext)
        assert not self._has_dh_ciphers(ssl_context)
        assert not self._has_weak_digest_ciphers(ssl_context)

    def test_tlsseclevel1adapter(self):
        adapter = TLSSecLevel1Adapter()
        ssl_context = adapter.poolmanager.connection_pool_kw.get("ssl_context")
        assert isinstance(ssl_context, SSLContext)
        assert self._has_dh_ciphers(ssl_context)
        assert self._has_weak_digest_ciphers(ssl_context)

    @pytest.mark.parametrize("proxy", ["http", "socks4", "socks5"])
    def test_proxymanager_ssl_context(self, proxy: str):
        adapter = SSLContextAdapter()
        proxymanager = adapter.proxy_manager_for(f"{proxy}://")
        ssl_context_poolmanager = adapter.poolmanager.connection_pool_kw.get("ssl_context")
        ssl_context_proxymanager = proxymanager.connection_pool_kw.get("ssl_context")
        assert ssl_context_poolmanager is ssl_context_proxymanager


class TestHTTPSessionVerifyAndCustomSSLContext:
    @pytest.fixture()
    def adapter(self, session: Streamlink):
        # The http-disable-dh session option mounts the TLSNoDHAdapter with a custom SSLContext
        session.set_option("http-disable-dh", True)

        adapter = session.http.adapters.get("https://")
        assert isinstance(adapter, TLSNoDHAdapter)

        return adapter

    @pytest.fixture(autouse=True)
    def _fake_request(self, monkeypatch: pytest.MonkeyPatch, session: Streamlink, adapter: HTTPAdapter):
        class FakeHTTPResponse(HTTPResponse):
            def stream(self, *_, **__):
                yield b"mocked"

        # Can't use requests_mock here, as it monkeypatches the adapter's send() method, which we want to test
        req = requests.PreparedRequest()
        resp = FakeHTTPResponse(status=200)
        # noinspection PyTypeChecker
        response = adapter.build_response(req, resp)
        monkeypatch.setattr("requests.adapters.HTTPAdapter.send", Mock(return_value=response))

        assert session.http.get("https://mocked/").text == "mocked"

    @pytest.mark.parametrize(
        ("session", "check_hostname", "verify_mode"),
        [
            pytest.param({"http-ssl-verify": True}, True, ssl.CERT_REQUIRED, id="verify"),
            pytest.param({"http-ssl-verify": False}, False, ssl.CERT_NONE, id="no-verify"),
        ],
        indirect=["session"],
    )
    def test_ssl_context_attributes(
        self,
        session: Streamlink,
        adapter: HTTPAdapter,
        check_hostname: bool,
        verify_mode: ssl.VerifyMode,
    ):
        assert session.http.verify is session.get_option("http-ssl-verify")

        ssl_context = adapter.poolmanager.connection_pool_kw.get("ssl_context")
        assert isinstance(ssl_context, SSLContext)
        assert ssl_context.check_hostname is check_hostname
        assert ssl_context.verify_mode is verify_mode



from streamlink.stream.wrappers import StreamIOIterWrapper


class TestPluginStream:
    def test_iter(self):
        def generator():
            yield b"1" * 8192
            yield b"2" * 4096
            yield b"3" * 2048

        fd = StreamIOIterWrapper(generator())
        assert fd.read(4096) == b"1" * 4096
        assert fd.read(2048) == b"1" * 2048
        assert fd.read(2048) == b"1" * 2048
        assert fd.read(1) == b"2"
        assert fd.read(4095) == b"2" * 4095
        assert fd.read(1536) == b"3" * 1536
        assert fd.read() == b"3" * 512



from __future__ import annotations

import math
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

import pytest

from streamlink.utils.dataclass import FormattedDataclass
from streamlink.utils.times import fromtimestamp


if TYPE_CHECKING:
    from datetime import datetime, timedelta


@pytest.fixture(scope="module")
def dataclass_hierarchy() -> tuple[type, type, type]:
    @dataclass(kw_only=True)
    class Foo(
        metaclass=FormattedDataclass,
        formatters={
            str: lambda s: f"{s.upper()!r}",
        },
    ):
        foo: str
        bar: int
        baz: float
        qux: str = field(repr=False)
        abc: datetime = field(default=fromtimestamp(0.5))
        xyz: timedelta = field(default=fromtimestamp(1.5) - fromtimestamp(0))

    @dataclass(kw_only=True)
    class Bar(
        Foo,
        metaclass=FormattedDataclass,
        formatters={
            float: lambda x: f"{(x * 2.0):.3f}",
        },
        extra=["oof"],
    ):
        @property
        def oof(self) -> str:
            return self.foo[::-1]

    @dataclass(kw_only=True)
    class Baz(Bar, metaclass=FormattedDataclass, extra=["barbar"]):
        @property
        def barbar(self) -> int:
            return self.bar * self.bar

    return Foo, Bar, Baz


@pytest.fixture(scope="module")
def baseclass(dataclass_hierarchy: tuple[type, type, type]) -> type:
    return dataclass_hierarchy[0]


@pytest.fixture(scope="module")
def subclass(dataclass_hierarchy: tuple[type, type, type]) -> type:
    return dataclass_hierarchy[1]


@pytest.fixture(scope="module")
def subsubclass(dataclass_hierarchy: tuple[type, type, type]) -> type:
    return dataclass_hierarchy[2]


@pytest.mark.parametrize(
    ("fixture", "expected"),
    [
        pytest.param(
            "baseclass",
            "Foo(foo='FOO', bar=123, baz=3.142, abc=1970-01-01T00:00:00.500000Z, xyz=0:00:01.500000)",
            id="baseclass",
        ),
        pytest.param(
            "subclass",
            "Bar(foo='FOO', bar=123, baz=6.283, abc=1970-01-01T00:00:00.500000Z, xyz=0:00:01.500000, oof='OOF')",
            id="subclass",
        ),
        pytest.param(
            "subsubclass",
            "Baz(foo='FOO', bar=123, baz=6.283, abc=1970-01-01T00:00:00.500000Z, xyz=0:00:01.500000, oof='OOF', barbar=15129)",
            id="subsubclass",
        ),
    ],
)
def test_serialize(request: pytest.FixtureRequest, fixture: str, expected: str):
    dc = request.getfixturevalue(fixture)
    item = dc(
        foo="foo",
        bar=123,
        baz=math.pi,
        qux="qux",
    )
    assert str(item) == repr(item) == expected



from unittest.mock import Mock

import pytest

import streamlink_cli.main


@pytest.mark.parametrize(
    ("argv", "latest", "aborted", "exit_code"),
    [
        pytest.param(
            ["--version-check"],
            True,
            True,
            130,
            id="aborted",
        ),
        pytest.param(
            ["--version-check"],
            True,
            False,
            0,
            id="latest",
        ),
        pytest.param(
            ["--version-check"],
            False,
            False,
            1,
            id="outdated",
        ),
        pytest.param(
            ["--auto-version-check=True"],
            True,
            False,
            0,
            id="auto-latest",
        ),
        pytest.param(
            ["--auto-version-check=True"],
            False,
            False,
            0,
            id="auto-outdated",
        ),
    ],
    indirect=["argv"],
)
def test_version_check(monkeypatch: pytest.MonkeyPatch, argv: list, latest: bool, aborted: bool, exit_code: int):
    mock_check_version = Mock(return_value=latest, side_effect=KeyboardInterrupt if aborted else None)
    monkeypatch.setattr("streamlink_cli.main.check_version", mock_check_version)

    with pytest.raises(SystemExit) as exc_info:
        streamlink_cli.main.main()
    assert exc_info.value.code == exit_code
    assert mock_check_version.call_count == 1



from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING
from unittest.mock import Mock, call

import pytest

from streamlink.stream.file import FileStream


if TYPE_CHECKING:
    from streamlink import Streamlink


class TestFileStream:
    @pytest.fixture()
    def mock_open(self, monkeypatch: pytest.MonkeyPatch) -> Mock:
        mock_open = Mock(return_value=BytesIO(b"foo"))
        monkeypatch.setattr("pathlib.Path.open", mock_open)

        return mock_open

    def test_no_path_no_fileobj(self, session: Streamlink):
        with pytest.raises(ValueError, match=r"^path or fileobj must be set$"):
            FileStream(session)

    def test_open_path(self, session: Streamlink, mock_open: Mock):
        stream = FileStream(session, path="/test/path")
        streamio = stream.open()
        assert mock_open.call_args_list == [call("rb")]
        assert streamio.read() == b"foo"

        streamio.close()

    def test_open_fileobj(self, session: Streamlink, mock_open: Mock):
        fileobj = BytesIO(b"bar")
        stream = FileStream(session, fileobj=fileobj)
        streamio = stream.open()
        assert streamio is fileobj
        assert mock_open.call_args_list == []
        assert streamio.read() == b"bar"

        streamio.close()



from __future__ import annotations

from contextlib import nullcontext
from typing import TYPE_CHECKING

import pytest

from streamlink.utils.args import boolean, comma_list, comma_list_filter, filesize, keyvalue, num


if TYPE_CHECKING:
    from collections.abc import Sequence


does_not_raise = nullcontext()


@pytest.mark.parametrize(
    ("value", "expected", "raises"),
    [
        ("1", True, does_not_raise),
        ("on", True, does_not_raise),
        ("true", True, does_not_raise),
        ("yes", True, does_not_raise),
        ("YES", True, does_not_raise),
        ("0", False, does_not_raise),
        ("off", False, does_not_raise),
        ("false", False, does_not_raise),
        ("no", False, does_not_raise),
        ("NO", False, does_not_raise),
        ("invalid", None, pytest.raises(ValueError, match=r"invalid is not one of .+")),
        ("2", None, pytest.raises(ValueError, match=r"2 is not one of .+")),
    ],
)
def test_boolean(value: str, expected: bool, raises: nullcontext):
    with raises:
        assert boolean(value) is expected


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        pytest.param("foo", ["foo"], id="single item"),
        pytest.param("foo.bar,example.com", ["foo.bar", "example.com"], id="separator"),
        pytest.param("/var/run/foo,/var/run/bar", ["/var/run/foo", "/var/run/bar"], id="paths"),
        pytest.param("foo bar,baz", ["foo bar", "baz"], id="whitespace"),
    ],
)
def test_comma_list(value: str, expected: list[str]):
    assert comma_list(value) == expected


@pytest.mark.parametrize(
    ("options", "value", "expected"),
    [
        pytest.param(
            {"acceptable": ["foo", "bar"]},
            "foo,bar,baz,qux",
            ["foo", "bar"],
            id="superset",
        ),
        pytest.param(
            {"acceptable": ["foo", "bar", "baz"]},
            "foo,bar",
            ["foo", "bar"],
            id="subset",
        ),
        pytest.param(
            {"acceptable": ["foo", "bar"], "unique": True},
            "foo,bar,baz,foo,bar,baz",
            ["bar", "foo"],
            id="unique",
        ),
    ],
)
def test_comma_list_filter(options: dict, value: str, expected: list[str]):
    func = comma_list_filter(**options)
    assert func(value) == expected


def test_comma_list_filter_hashable():
    assert hash(comma_list_filter(["1", "2"])) != hash(comma_list_filter(["1", "2", "3"]))
    assert hash(comma_list_filter(["1", "2"], unique=True)) == hash(comma_list_filter(["1", "2"], unique=True))
    assert hash(comma_list_filter(["1", "2"], unique=True)) != hash(comma_list_filter(["1", "2"], unique=False))


@pytest.mark.parametrize(
    ("value", "expected", "raises"),
    [
        ("12345", 12345, does_not_raise),
        ("123.45", int(123.45), does_not_raise),
        ("1KB", 1 * 2**10, does_not_raise),
        ("123kB", 123 * 2**10, does_not_raise),
        ("123.45kB", int(123.45 * 2**10), does_not_raise),
        ("1mb", 1 * 2**20, does_not_raise),
        ("123k", 123 * 2**10, does_not_raise),
        ("123M", 123 * 2**20, does_not_raise),
        ("123.45M", int(123.45 * 2**20), does_not_raise),
        ("  123.45MB  ", int(123.45 * 2**20), does_not_raise),
        ("FOO", None, pytest.raises(ValueError, match=r"^Invalid file size format$")),
        ("0", None, pytest.raises(ValueError, match=r"^int value must be >=1, but is 0$")),
        ("0.00000", None, pytest.raises(ValueError, match=r"^int value must be >=1, but is 0$")),
    ],
)
def test_filesize(value: str, expected: int, raises: nullcontext):
    with raises:
        assert filesize(value) == expected


@pytest.mark.parametrize(
    ("value", "expected", "raises"),
    [
        pytest.param(
            "X-Forwarded-For=127.0.0.1",
            ("X-Forwarded-For", "127.0.0.1"),
            does_not_raise,
            id="separator",
        ),
        pytest.param(
            "foo=bar=baz",
            ("foo", "bar=baz"),
            does_not_raise,
            id="value with separator",
        ),
        pytest.param(
            "foo=",
            ("foo", ""),
            does_not_raise,
            id="missing value",
        ),
        pytest.param(
            "  foo  =  bar  ",
            ("foo", "bar  "),
            does_not_raise,
            id="whitespace",
        ),
        pytest.param(
            "User-Agent=Mozilla/5.0 (X11; Linux x86_64; rv:60.0) Gecko/20100101 Firefox/60.0",
            ("User-Agent", "Mozilla/5.0 (X11; Linux x86_64; rv:60.0) Gecko/20100101 Firefox/60.0"),
            does_not_raise,
            id="user-agent",
        ),
        pytest.param(
            "127.0.0.1",
            None,
            pytest.raises(ValueError, match=r"^Invalid key=value format$"),
            id="invalid format",
        ),
        pytest.param(
            "=value",
            None,
            pytest.raises(ValueError, match=r"^Invalid key=value format$"),
            id="missing key",
        ),
    ],
)
def test_keyvalue(value: str, expected: Sequence[str], raises: nullcontext):
    with raises:
        assert keyvalue(value) == expected


class TestNum:
    @pytest.mark.parametrize(
        ("numtype", "value", "expected", "raises"),
        [
            (int, 123, 123, does_not_raise),
            (float, 123, 123.0, does_not_raise),
            (int, 123.456, 123, does_not_raise),
            (int, "123", 123, does_not_raise),
            (int, "-123", -123, does_not_raise),
            (float, "123.456", 123.456, does_not_raise),
            (float, "3.1415e2", 314.15, does_not_raise),
            (int, "", None, pytest.raises(ValueError, match=r"^invalid literal for int\(\) with base 10:")),
            (int, ".", None, pytest.raises(ValueError, match=r"^invalid literal for int\(\) with base 10:")),
            (int, "-", None, pytest.raises(ValueError, match=r"^invalid literal for int\(\) with base 10:")),
            (int, "foo", None, pytest.raises(ValueError, match=r"^invalid literal for int\(\) with base 10:")),
            (float, "", None, pytest.raises(ValueError, match=r"^could not convert string to float:")),
            (float, ".", None, pytest.raises(ValueError, match=r"^could not convert string to float:")),
            (float, "-", None, pytest.raises(ValueError, match=r"^could not convert string to float:")),
            (float, "foo", None, pytest.raises(ValueError, match=r"^could not convert string to float:")),
        ],
    )
    def test_numtype(self, numtype: type[float], value: float, expected: float, raises: nullcontext):
        func = num(numtype)
        assert func.__name__ == numtype.__name__
        with raises:
            result = func(value)
            assert type(result) is numtype
            assert result == expected

    @pytest.mark.parametrize(
        ("operators", "value", "raises"),
        [
            ({"ge": 1}, 1, does_not_raise),
            ({"ge": 0}, 1, does_not_raise),
            ({"gt": 0}, 1, does_not_raise),
            ({"le": 1}, 1, does_not_raise),
            ({"le": 2}, 1, does_not_raise),
            ({"lt": 2}, 1, does_not_raise),
            ({"ge": 1, "gt": 0, "le": 1, "lt": 2}, 1, does_not_raise),
            ({"ge": 2}, 1, pytest.raises(ValueError, match=r"^int value must be >=2, but is 1$")),
            ({"gt": 1}, 1, pytest.raises(ValueError, match=r"^int value must be >1, but is 1$")),
            ({"le": 0}, 1, pytest.raises(ValueError, match=r"^int value must be <=0, but is 1$")),
            ({"lt": 1}, 1, pytest.raises(ValueError, match=r"^int value must be <1, but is 1$")),
            ({"ge": 1, "gt": 0, "le": 0, "lt": 2}, 1, pytest.raises(ValueError, match=r"^int value must be <=0, but is 1$")),
        ],
    )
    def test_operator(self, operators: dict, value: float, raises: nullcontext):
        with raises:
            assert num(int, **operators)(value) == value

    def test_hashable(self):
        assert hash(num(int, ge=1)) == hash(num(int, ge=1))
        assert hash(num(float, ge=1.0)) == hash(num(float, ge=1.0))
        assert hash(num(int, ge=1)) != hash(num(int, gt=1))
        assert hash(num(int, ge=1)) != hash(num(int, le=1))
        assert hash(num(int, ge=1)) != hash(num(int, lt=1))
        assert hash(num(int, ge=1)) != hash(num(float, ge=1.0))



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING
from unittest.mock import patch

import freezegun
import pytest

from streamlink.plugins.filmon import Filmon, FilmOnAPI, FilmOnHLS
from tests.plugins import PluginCanHandleUrl


if TYPE_CHECKING:
    from streamlink import Streamlink


class TestPluginCanHandleUrlFilmon(PluginCanHandleUrl):
    __plugin__ = Filmon

    should_match = [
        "http://www.filmon.tv/channel/grandstand-show",
        "http://www.filmon.tv/index/popout?channel_id=5510&quality=low",
        "http://www.filmon.tv/tv/channel/export?channel_id=5510&autoPlay=1",
        "http://www.filmon.tv/tv/channel/grandstand-show",
        "http://www.filmon.tv/tv/channel-4",
        "https://www.filmon.com/tv/bbc-news",
        "https://www.filmon.tv/tv/55",
        "http://www.filmon.tv/vod/view/10250-0-crime-boss",
        "http://www.filmon.tv/group/comedy",
    ]

    should_match_groups = [
        ("http://www.filmon.tv/channel/grandstand-show", (None, "grandstand-show", None)),
        ("http://www.filmon.tv/index/popout?channel_id=5510&quality=low", (None, "5510", None)),
        ("http://www.filmon.tv/tv/channel/export?channel_id=5510&autoPlay=1", (None, "5510", None)),
        ("http://www.filmon.tv/tv/channel/grandstand-show", (None, "grandstand-show", None)),
        ("https://www.filmon.com/tv/bbc-news", (None, "bbc-news", None)),
        ("https://www.filmon.com/tv/channel-4", (None, "channel-4", None)),
        ("https://www.filmon.tv/tv/55", (None, "55", None)),
        ("http://www.filmon.tv/group/comedy", ("group/", "comedy", None)),
        ("http://www.filmon.tv/vod/view/10250-0-crime-boss", (None, None, "10250-0-crime-boss")),
        ("http://www.filmon.tv/vod/view/10250-0-crime-boss/extra", (None, None, "10250-0-crime-boss")),
        ("http://www.filmon.tv/vod/view/10250-0-crime-boss?extra", (None, None, "10250-0-crime-boss")),
        ("http://www.filmon.tv/vod/view/10250-0-crime-boss&extra", (None, None, "10250-0-crime-boss")),
    ]


@pytest.fixture()
def filmonhls(session: Streamlink):
    with (
        freezegun.freeze_time("2000-01-01T00:00:00Z"),
        patch("streamlink.plugins.filmon.FilmOnHLS._get_stream_data", return_value=[]),
    ):
        api = FilmOnAPI(session)
        yield FilmOnHLS(session, "http://fake/one.m3u8", api=api, channel="test")


def test_filmonhls_to_url(filmonhls):
    filmonhls.watch_timeout = datetime(2000, 1, 1, 0, 0, 0, 0, timezone.utc).timestamp()
    assert filmonhls.to_url() == "http://fake/one.m3u8"


def test_filmonhls_to_url_updated(filmonhls):
    filmonhls.watch_timeout = datetime(1999, 12, 31, 23, 59, 59, 9999, timezone.utc).timestamp()

    filmonhls._get_stream_data.return_value = [
        ("high", "http://fake/two.m3u8", datetime(2000, 1, 1, 0, 0, 0, 0, timezone.utc).timestamp()),
    ]
    assert filmonhls.to_url() == "http://fake/two.m3u8"

    filmonhls.watch_timeout = datetime(1999, 12, 31, 23, 59, 59, 9999, timezone.utc).timestamp()
    filmonhls._get_stream_data.return_value = [
        ("high", "http://another-fake/three.m3u8", datetime(2000, 1, 1, 0, 0, 0, 0, timezone.utc).timestamp()),
    ]
    assert filmonhls.to_url() == "http://fake/three.m3u8"


def test_filmonhls_to_url_missing_quality(filmonhls):
    filmonhls.watch_timeout = datetime(1999, 12, 31, 23, 59, 59, 9999, timezone.utc).timestamp()

    filmonhls._get_stream_data.return_value = [
        ("low", "http://fake/two.m3u8", datetime(2000, 1, 1, 0, 0, 0, 0, timezone.utc).timestamp()),
    ]
    with pytest.raises(TypeError) as cm:
        filmonhls.to_url()
    assert str(cm.value) == "Stream has expired and cannot be translated to a URL"



from streamlink.plugins.yupptv import YuppTV
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlYuppTV(PluginCanHandleUrl):
    __plugin__ = YuppTV

    should_match = [
        "https://www.yupptv.com/online-tv/wion/live",
        "https://www.yupptv.com/online-tv/mirror-now/live",
        "https://www.yupptv.com/channels/etv-telugu/live",
        "https://www.yupptv.com/channels/india-today-news/news/25326023/15-jun-2018",
    ]



from unittest.mock import ANY, AsyncMock, call

import pytest

from tests.webbrowser.cdp import FakeWebsocketConnection


@pytest.fixture()
def websocket_connection(monkeypatch: pytest.MonkeyPatch):
    fake_websocket_connection = FakeWebsocketConnection()
    mock_connect_websocket_url = AsyncMock(return_value=fake_websocket_connection)
    monkeypatch.setattr("streamlink.webbrowser.cdp.connection.connect_websocket_url", mock_connect_websocket_url)

    try:
        yield fake_websocket_connection
    finally:
        assert fake_websocket_connection.closed
        assert mock_connect_websocket_url.call_args_list == [call(ANY, "ws://localhost:1234/fake", max_message_size=2**24)]



from streamlink.plugins.earthcam import EarthCam
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlEarthCam(PluginCanHandleUrl):
    __plugin__ = EarthCam

    should_match = [
        "https://www.earthcam.com/usa/newyork/timessquare/?cam=tsstreet",
        "https://www.earthcam.com/usa/newyork/timessquare/?cam=gts1",
    ]



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

import tests.plugin
from streamlink.api import streams


if TYPE_CHECKING:
    from streamlink import Streamlink


class TestStreamlinkAPI:
    @pytest.fixture(autouse=True)
    def _session(self, monkeypatch: pytest.MonkeyPatch, session: Streamlink):
        monkeypatch.setattr("streamlink.api.Streamlink", lambda: session)
        session.plugins.load_path(tests.plugin.__path__[0])

    def test_find_test_plugin(self):
        assert "hls" in streams("test.se")

    def test_no_streams_exception(self):
        assert streams("test.se/NoStreamsError") == {}

    def test_no_streams(self):
        assert streams("test.se/empty") == {}

    def test_stream_type_filter(self):
        stream_types = ["hls"]
        available_streams = streams("test.se", stream_types=stream_types)
        assert "hls" in available_streams
        assert "test" not in available_streams
        assert "http" not in available_streams

    def test_stream_type_wildcard(self):
        stream_types = ["hls", "*"]
        available_streams = streams("test.se", stream_types=stream_types)
        assert "hls" in available_streams
        assert "test" in available_streams
        assert "http" in available_streams



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import Mock, call

import pytest

import streamlink_cli.main


if TYPE_CHECKING:
    from streamlink import Streamlink


@pytest.fixture(autouse=True)
def _cwd(monkeypatch: pytest.MonkeyPatch, tmp_path: Path):
    monkeypatch.chdir(tmp_path)
    Path(tmp_path, "DEFAULT").mkdir()
    Path(tmp_path, "custom1").mkdir()


@pytest.fixture(autouse=True)
def _run(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr("streamlink_cli.main.run", Mock(return_value=0))


@pytest.fixture()
def plugin_dirs(request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch):
    value = getattr(request, "param", None)
    plugin_dirs = value if value is not None else [Path("DEFAULT"), Path("DOESNOTEXIST")]
    monkeypatch.setattr("streamlink_cli.main.PLUGIN_DIRS", plugin_dirs)


@pytest.fixture()
def mock_load_path(request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch, session: Streamlink):
    mock_load_path = Mock()
    monkeypatch.setattr(session.plugins, "load_path", mock_load_path)
    return mock_load_path


@pytest.mark.parametrize(
    ("argv", "plugin_dirs", "expected", "log"),
    [
        pytest.param(
            [],
            None,
            [
                call(Path("DEFAULT")),
            ],
            [],
            id="default",
        ),
        pytest.param(
            ["--no-plugin-sideloading"],
            None,
            [],
            [],
            id="no-plugin-sideloading",
        ),
        pytest.param(
            ["--plugin-dir=custom1", "--plugin-dir=custom2"],
            None,
            [
                call(Path("DEFAULT")),
                call(Path("custom1")),
            ],
            [
                ("cli", "warning", "Plugin path custom2 does not exist or is not a directory!"),
            ],
            id="custom-paths",
        ),
        pytest.param(
            ["--plugin-dirs=custom1,custom2"],
            None,
            [
                call(Path("DEFAULT")),
                call(Path("custom1")),
            ],
            [
                ("cli", "warning", "Plugin path custom2 does not exist or is not a directory!"),
            ],
            id="custom-paths-deprecated",
        ),
        pytest.param(
            ["--no-plugin-sideloading", "--plugin-dir=custom1", "--plugin-dir=custom2"],
            None,
            [
                call(Path("custom1")),
            ],
            [
                ("cli", "warning", "Plugin path custom2 does not exist or is not a directory!"),
            ],
            id="custom-paths-without-sideloading",
        ),
    ],
    indirect=["argv", "plugin_dirs"],
)
def test_setup_plugins(
    caplog: pytest.LogCaptureFixture,
    argv: list,
    plugin_dirs: list,
    mock_load_path: Mock,
    expected: list[str],
    log: list,
):
    caplog.set_level(1, "streamlink")

    with pytest.raises(SystemExit):
        streamlink_cli.main.main()

    assert mock_load_path.call_args_list == expected
    assert [(record.name, record.levelname, record.message) for record in caplog.records] == log



from streamlink.plugins.bloomberg import Bloomberg
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlBloomberg(PluginCanHandleUrl):
    __plugin__ = Bloomberg

    should_match_groups = [
        (
            ("live", "https://www.bloomberg.com/live"),
            {},
        ),
        (
            ("live", "https://www.bloomberg.com/live/europe"),
            {"channel": "europe"},
        ),
        (
            ("live", "https://www.bloomberg.com/live/us"),
            {"channel": "us"},
        ),
        (
            ("vod", "https://www.bloomberg.com/news/videos/2022-08-10/-bloomberg-surveillance-early-edition-full-08-10-22"),
            {},
        ),
    ]

    should_not_match = [
        "https://www.bloomberg.com/politics/articles/2017-04-17/french-race-up-for-grabs-days-before-voters-cast-first-ballots",
    ]



from __future__ import annotations

import sys

# noinspection PyProtectedMember
from ctypes import Structure, _SimpleCData  # noqa: PLC2701
from ctypes.wintypes import DWORD, WCHAR, WORD
from types import ModuleType
from typing import TYPE_CHECKING, ClassVar, Generic, TypeVar
from unittest.mock import ANY, Mock, call

import pytest

from streamlink_cli.console.windows import COORD, WindowsConsole


if TYPE_CHECKING:
    from io import TextIOWrapper


_TCTypesType = TypeVar("_TCTypesType")


class _CTypesComparable(Generic[_TCTypesType]):
    """Allow comparing ctypes data types with built-in types for equality"""

    _type: ClassVar[type]

    def __init__(self, data: _TCTypesType):
        self.data: tuple = self._get_data(data)

    def _get_data(self, data: _TCTypesType) -> tuple:  # pragma: no cover
        raise NotImplementedError

    def __eq__(self, other):
        if isinstance(other, self._type):
            return self.__eq__(type(self)(other))
        if not isinstance(other, type(self)):  # pragma: no cover
            return False

        return self.data == other.data

    def __hash__(self):  # pragma: no cover
        return super().__hash__()


class EqSimpleCData(_CTypesComparable[_SimpleCData]):
    _type: ClassVar = _SimpleCData

    def _get_data(self, data: _SimpleCData) -> tuple:
        # noinspection PyProtectedMember
        return data._type_, data.value  # type: ignore[attr-defined, ty:unresolved-attribute]


class EqStructure(_CTypesComparable[Structure]):
    _type: ClassVar = Structure

    def _get_data(self, data: Structure) -> tuple:
        # noinspection PyProtectedMember
        return tuple(getattr(data, item) for (item, *_) in data._fields_)


@pytest.fixture(autouse=True)
def _mock_byref(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr("streamlink_cli.console.windows.byref", Mock(side_effect=lambda obj, *_, **__: obj))


@pytest.fixture(autouse=True)
def mock_windll(monkeypatch: pytest.MonkeyPatch):
    mock_windll = Mock()
    monkeypatch.setattr("ctypes.windll", mock_windll, raising=False)

    return mock_windll


def test_no_windll(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setitem(sys.modules, "ctypes", ModuleType("ctypes"))
    assert WindowsConsole() is None


@pytest.mark.parametrize(
    ("method", "function"),
    [
        ("get_std_handle", "GetStdHandle"),
        ("get_console_mode", "GetConsoleMode"),
        ("get_console_screen_buffer_info", "GetConsoleScreenBufferInfo"),
        ("set_console_cursor_position", "SetConsoleCursorPosition"),
        ("fill_console_output_attribute", "FillConsoleOutputAttribute"),
        ("fill_console_output_character_w", "FillConsoleOutputCharacterW"),
    ],
)
def test_functions(mock_windll: Mock, method: str, function: str):
    windows_console = WindowsConsole()
    assert isinstance(windows_console, WindowsConsole)
    assert getattr(windows_console, method).method is getattr(mock_windll.kernel32, function)


def test_call_success_error(monkeypatch: pytest.MonkeyPatch, mock_windll: Mock):
    windows_console = WindowsConsole()
    monkeypatch.setattr(windows_console.set_console_cursor_position, "method", Mock(return_value=False))
    mock_windll.kernel32.GetLastError.return_value = 87

    with pytest.raises(OSError) as exc_info:  # noqa: PT011
        windows_console.set_console_cursor_position(123, 456)  # type: ignore[arg-type, ty:invalid-argument-type]
    assert str(exc_info.value) == "Error while calling kernel32.SetConsoleCursorPosition (last_error=0x57)"


@pytest.mark.parametrize(
    ("stream", "expected"),
    [
        pytest.param(None, -11, id="None"),
        pytest.param(sys.stdout, -11, id="stdout"),
        pytest.param(sys.stderr, -12, id="stderr"),
    ],
)
def test_std_handle(mock_windll: Mock, stream: TextIOWrapper | None, expected: int):
    windows_console = WindowsConsole(stream)
    assert mock_windll.kernel32.GetStdHandle.call_args_list == [call(expected)]
    assert windows_console.handle is mock_windll.kernel32.GetStdHandle.return_value


@pytest.mark.parametrize(
    ("value", "success", "expected"),
    [
        pytest.param(0, False, False, id="error"),
        pytest.param(3, True, False, id="no-virtual-terminal-processing"),
        pytest.param(7, True, True, id="virtual-terminal-processing"),
    ],
)
def test_supports_virtual_terminal_processing(mock_windll: Mock, value: int, success: bool, expected: bool):
    def fake_get_console_mode(_handle, mode):
        mode.value = value

        return success

    mock_windll.kernel32.GetConsoleMode.side_effect = fake_get_console_mode
    mock_windll.kernel32.GetLastError.return_value = 87

    windows_console = WindowsConsole()
    assert windows_console.supports_virtual_terminal_processing() == expected
    assert mock_windll.kernel32.GetConsoleMode.call_args_list == [call(mock_windll.kernel32.GetStdHandle.return_value, ANY)]


def test_clear_line(mock_windll: Mock):
    def fake_get_console_screen_buffer_info(_handle, console_screen_buffer_info):
        console_screen_buffer_info.dwSize.X = 144
        console_screen_buffer_info.dwSize.Y = 42
        console_screen_buffer_info.dwCursorPosition.X = 20
        console_screen_buffer_info.dwCursorPosition.Y = 15
        console_screen_buffer_info.wAttributes = 0

        return True

    mock_windll.kernel32.GetConsoleScreenBufferInfo.side_effect = fake_get_console_screen_buffer_info

    windows_console = WindowsConsole()
    windows_console.clear_line()
    assert mock_windll.kernel32.GetConsoleScreenBufferInfo.call_args_list == [
        call(mock_windll.kernel32.GetStdHandle.return_value, ANY),
    ]
    assert mock_windll.kernel32.FillConsoleOutputCharacterW.call_args_list == [
        call(
            mock_windll.kernel32.GetStdHandle.return_value,
            EqSimpleCData(WCHAR(" ")),
            EqSimpleCData(DWORD(144)),
            EqStructure(COORD(0, 15)),
            EqSimpleCData(DWORD(0)),
        ),
    ]
    assert mock_windll.kernel32.FillConsoleOutputAttribute.call_args_list == [
        call(
            mock_windll.kernel32.GetStdHandle.return_value,
            EqSimpleCData(WORD(0)),
            EqSimpleCData(DWORD(144)),
            EqStructure(COORD(0, 15)),
            EqSimpleCData(DWORD(0)),
        ),
    ]
    assert mock_windll.kernel32.SetConsoleCursorPosition.call_args_list == [
        call(
            mock_windll.kernel32.GetStdHandle.return_value,
            EqStructure(COORD(0, 15)),
        ),
    ]



from urllib.parse import quote

import pytest

from streamlink.utils.url import absolute_url, is_insecure_scheme, prepend_www, update_qsd, update_scheme, url_concat, url_equal


@pytest.mark.parametrize(
    ("baseurl", "url", "expected"),
    [
        ("http://test.se", "/test", "http://test.se/test"),
        ("http://test.se", "http/test.se/test", "http://test.se/http/test.se/test"),
        ("http://test.se", "http://test2.se/test", "http://test2.se/test"),
    ],
)
def test_absolute_url(baseurl, url, expected):
    assert expected == absolute_url(baseurl, url)


@pytest.mark.parametrize(
    ("url", "expected"),
    [
        ("http://test.se/test", "http://www.test.se/test"),
        ("http://www.test.se", "http://www.test.se"),
    ],
)
def test_prepend_www(url, expected):
    assert expected == prepend_www(url)


@pytest.mark.parametrize(
    ("current_scheme", "target_scheme", "expected"),
    [
        ("http", "", False),
        ("https", "", False),
        ("file", "", False),
        ("http", "http", False),
        ("http", "https", False),
        ("https", "https", False),
        ("https", "http", True),
        ("file", "file", False),
        ("http", "file", True),
        ("https", "file", True),
    ],
)
def test_is_insecure_scheme(current_scheme: str, target_scheme: str, expected: bool):
    assert is_insecure_scheme(current_scheme, target_scheme) == expected


@pytest.mark.parametrize(
    ("args", "expected"),
    [
        pytest.param(
            ("https://other.com/bar", "http://example.com/foo"),
            "https://example.com/foo",
            id="override-target-http-scheme",
        ),
        pytest.param(
            ("http://other.com/bar", "https://example.com/foo"),
            "http://example.com/foo",
            id="override-target-https-scheme",
        ),
        pytest.param(
            ("http://other.com/bar", "https://example.com/foo", False),
            "https://example.com/foo",
            id="keep-target-https-scheme-no-force",
        ),
        pytest.param(
            ("https://other.com/bar", "http://example.com/foo", False),
            "http://example.com/foo",
            id="keep-target-http-scheme-no-force",
        ),
        pytest.param(
            ("https://other.com/bar", "//example.com/foo"),
            "https://example.com/foo",
            id="add-current-scheme-to-schemeless-target",
        ),
        pytest.param(
            ("https://other.com/bar", "//example.com/foo", False),
            "https://example.com/foo",
            id="add-current-scheme-to-schemeless-target-no-force",
        ),
        pytest.param(
            ("https://other.com/bar", "example.com/foo"),
            "https://example.com/foo",
            id="add-current-scheme-to-target-string",
        ),
        pytest.param(
            ("https://other.com/bar", "example.com/foo", False),
            "https://example.com/foo",
            id="add-current-scheme-to-target-string-no-force",
        ),
        pytest.param(
            ("http://", "127.0.0.1:1234"),
            "http://127.0.0.1:1234",
            id="implicit-scheme-with-ipv4-port",
        ),
        pytest.param(
            ("http://", "foo.bar:1234"),
            "http://foo.bar:1234",
            id="implicit-scheme-with-hostname-port",
        ),
        pytest.param(
            ("foo.1+2-bar://baz", "FOO.1+2-BAR://qux"),
            "foo.1+2-bar://qux",
            id="weird-schemes",
        ),
        pytest.param(
            ("https://", "file://./foo", False),
            "file://./foo",
            id="keep-target",
        ),
        pytest.param(
            ("https://", "file:///foo", False),
            "file:///foo",
            id="keep-target-no-netloc",
        ),
    ],
)
def test_update_scheme(args, expected):
    assert update_scheme(*args) == expected


def test_url_equal():
    assert url_equal("http://test.com/test", "http://test.com/test")
    assert not url_equal("http://test.com/test", "http://test.com/test2")

    assert url_equal("http://test.com/test", "http://test.com/test2", ignore_path=True)
    assert url_equal("http://test.com/test", "https://test.com/test", ignore_scheme=True)
    assert not url_equal("http://test.com/test", "https://test.com/test")

    assert url_equal("http://test.com/test", "http://test.com/test#hello", ignore_fragment=True)
    assert url_equal("http://test.com/test", "http://test2.com/test", ignore_netloc=True)
    assert not url_equal("http://test.com/test", "http://test2.com/test1", ignore_netloc=True)


def test_url_concat():
    assert url_concat("http://test.se", "one", "two", "three") == "http://test.se/one/two/three"
    assert url_concat("http://test.se", "/one", "/two", "/three") == "http://test.se/one/two/three"
    assert url_concat("http://test.se/one", "../two", "three") == "http://test.se/two/three"
    assert url_concat("http://test.se/one", "../two", "../three") == "http://test.se/three"


def test_update_qsd():
    assert update_qsd("http://test.se?one=1&two=3", {"two": 2}) == "http://test.se?one=1&two=2"
    assert update_qsd("http://test.se?one=1&two=3", remove=["two"]) == "http://test.se?one=1"
    assert update_qsd("http://test.se?one=1&two=3", {"one": None}, remove="*") == "http://test.se?one=1"
    assert update_qsd("http://test.se", {"one": "", "two": ""}) == "http://test.se?one=&two=", "should add empty params"
    assert update_qsd("http://test.se?one=", {"one": None}) == "http://test.se?one=", "should leave empty params unchanged"
    assert update_qsd("http://test.se?one=", keep_blank_values=False) == "http://test.se", "should strip blank params"
    assert update_qsd("http://test.se?one=&two=", {"one": None}, keep_blank_values=False) == "http://test.se?one=", \
        "should leave one"  # fmt: skip
    assert update_qsd("http://test.se?&two=", {"one": ""}, keep_blank_values=False) == "http://test.se?one=", \
        "should set one blank"  # fmt: skip
    assert update_qsd("http://test.se?one=", {"two": 2}) == "http://test.se?one=&two=2"

    assert update_qsd("http://test.se?foo=%3F", {"bar": "!"}) == "http://test.se?foo=%3F&bar=%21", \
        "urlencode - encoded URL"  # fmt: skip
    assert update_qsd("http://test.se?foo=?", {"bar": "!"}) == "http://test.se?foo=%3F&bar=%21", \
        "urlencode - fix URL"  # fmt: skip
    assert update_qsd("http://test.se?foo=?", {"bar": "!"}, quote_via=lambda s, *_: s) == "http://test.se?foo=?&bar=!", \
        "urlencode - dummy quote method"  # fmt: skip
    assert update_qsd("http://test.se", {"foo": "/ "}) == "http://test.se?foo=%2F+", \
        "urlencode - default quote_plus"  # fmt: skip
    assert update_qsd("http://test.se", {"foo": "/ "}, safe="/", quote_via=quote) == "http://test.se?foo=/%20", \
        "urlencode - regular quote with reserved slash"  # fmt: skip
    assert update_qsd("http://test.se", {"foo": "/ "}, safe="", quote_via=quote) == "http://test.se?foo=%2F%20", \
        "urlencode - regular quote without reserved slash"  # fmt: skip



from contextlib import nullcontext
from pathlib import Path, PurePosixPath
from typing import TYPE_CHECKING
from unittest.mock import Mock, call

import pytest

from streamlink_cli.exceptions import StreamlinkCLIError
from streamlink_cli.main import check_file_output


if TYPE_CHECKING:
    _BasePath = PurePosixPath
else:
    _BasePath = type(PurePosixPath())


does_not_raise = nullcontext()


# Fake PurePosixPath, with a fake is_file() method which gets mocked in the path fixture down below:
# Can't override/extend the constructor with custom args/kwargs due to major code changes in py310
# which are incompatible with older versions of pathlib.PurePath
class _FakePath(_BasePath):
    @staticmethod
    def is_file():  # pragma: no cover
        return False


@pytest.fixture(autouse=True)
def _caplog(caplog: pytest.LogCaptureFixture):
    caplog.set_level(1, "streamlink.cli")


@pytest.fixture(autouse=True)
def path(monkeypatch: pytest.MonkeyPatch, request: pytest.FixtureRequest):
    param = getattr(request, "param", {})
    file = param.get("file", "file")
    realpath = param.get("realpath", "/path/to/file")
    exists = param.get("exists", False)

    monkeypatch.setattr(Path, "resolve", Mock(return_value=_FakePath(realpath)))
    monkeypatch.setattr(_FakePath, "is_file", Mock(return_value=exists))

    return Path(file)


@pytest.fixture()
def prompt(request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch):
    param = getattr(request, "param", {})
    ask = param.get("ask", "y")

    prompt = Mock(side_effect=ask) if isinstance(ask, Exception) else Mock(return_value=ask)
    monkeypatch.setattr("streamlink_cli.main.console", Mock(ask=prompt))

    return prompt


@pytest.mark.parametrize(
    ("path", "skip", "force", "raises", "log"),
    [
        pytest.param(
            {"exists": False},
            False,
            False,
            does_not_raise,
            [
                ("streamlink.cli", "info", "Writing output to\n/path/to/file"),
                ("streamlink.cli", "debug", "Checking file output"),
            ],
            id="does-not-exist",
        ),
        pytest.param(
            {"exists": True},
            False,
            True,
            does_not_raise,
            [
                ("streamlink.cli", "info", "Writing output to\n/path/to/file"),
                ("streamlink.cli", "debug", "Checking file output"),
            ],
            id="exists-force",
        ),
        pytest.param(
            {"exists": True},
            True,
            False,
            pytest.raises(StreamlinkCLIError),
            [
                ("streamlink.cli", "info", "Writing output to\n/path/to/file"),
                ("streamlink.cli", "debug", "Checking file output"),
                ("streamlink.cli", "error", "File file already exists"),
            ],
            id="exists-skip",
        ),
        pytest.param(
            {"exists": True},
            True,
            True,
            pytest.raises(StreamlinkCLIError),
            [
                ("streamlink.cli", "info", "Writing output to\n/path/to/file"),
                ("streamlink.cli", "debug", "Checking file output"),
                ("streamlink.cli", "error", "File file already exists"),
            ],
            id="exists-skip-and-force",
        ),
    ],
    indirect=["path"],
)
def test_exists(
    caplog: pytest.LogCaptureFixture,
    prompt: Mock,
    path: Path,
    skip: bool,
    force: bool,
    raises: nullcontext,
    log: list,
):
    with raises:
        output = check_file_output(path, skip, force)
        assert isinstance(output, _FakePath)
        assert output == PurePosixPath("/path/to/file")

    assert [(record.name, record.levelname, record.message) for record in caplog.records] == log
    assert prompt.call_args_list == []


@pytest.mark.parametrize("path", [pytest.param({"exists": True}, id="")], indirect=True)
@pytest.mark.parametrize(
    ("prompt", "exits", "log"),
    [
        pytest.param(
            {"ask": "y"},
            does_not_raise,
            [
                ("streamlink.cli", "info", "Writing output to\n/path/to/file"),
                ("streamlink.cli", "debug", "Checking file output"),
            ],
            id="yes",
        ),
        pytest.param(
            {"ask": "n"},
            pytest.raises(StreamlinkCLIError),
            [
                ("streamlink.cli", "info", "Writing output to\n/path/to/file"),
                ("streamlink.cli", "debug", "Checking file output"),
            ],
            id="no",
        ),
        pytest.param(
            {"ask": None},
            pytest.raises(StreamlinkCLIError),
            [
                ("streamlink.cli", "info", "Writing output to\n/path/to/file"),
                ("streamlink.cli", "debug", "Checking file output"),
            ],
            id="none",
        ),
        pytest.param(
            {"ask": OSError()},
            pytest.raises(StreamlinkCLIError),
            [
                ("streamlink.cli", "info", "Writing output to\n/path/to/file"),
                ("streamlink.cli", "debug", "Checking file output"),
                ("streamlink.cli", "error", "File file already exists, use --force to overwrite it or --skip this prompt"),
            ],
            id="oserror",
        ),
    ],
    indirect=["prompt"],
)
def test_prompt(
    caplog: pytest.LogCaptureFixture,
    path: Path,
    prompt: Mock,
    exits: nullcontext,
    log: list,
):
    with exits:
        output = check_file_output(path, False, False)
        assert isinstance(output, _FakePath)
        assert output == PurePosixPath("/path/to/file")

    assert [(record.name, record.levelname, record.message) for record in caplog.records] == log
    assert prompt.call_args_list == [call("File file already exists! Overwrite it? [y/N] ")]



from streamlink.plugins.ard_live import ARDLive
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlARDLive(PluginCanHandleUrl):
    __plugin__ = ARDLive

    should_match = [
        "https://daserste.de/live/index.html",
        "https://www.daserste.de/live/index.html",
    ]

    should_not_match = [
        "http://mediathek.daserste.de/live",
    ]



from __future__ import annotations

import subprocess
from typing import TYPE_CHECKING
from unittest.mock import Mock, call

import pytest

from streamlink.stream.ffmpegmux import FFMPEGMuxer, FFmpegVersionOutput


if TYPE_CHECKING:
    from streamlink import Streamlink


# noinspection PyProtectedMember
@pytest.fixture(autouse=True)
def _resolve_command_cache_clear():
    FFMPEGMuxer._resolve_command.cache_clear()
    yield
    FFMPEGMuxer._resolve_command.cache_clear()


@pytest.fixture(autouse=True)
def _logger(caplog: pytest.LogCaptureFixture):
    caplog.set_level(1, "streamlink")


@pytest.fixture()
def session(session: Streamlink):
    session.set_option("ffmpeg-no-validation", True)

    return session


class TestCommand:
    def test_cache(self, monkeypatch: pytest.MonkeyPatch, session: Streamlink):
        mock = Mock()
        monkeypatch.setattr("streamlink.stream.ffmpegmux.which", mock)

        mock.return_value = "some_value"
        assert FFMPEGMuxer.command(session) == "some_value"
        assert FFMPEGMuxer.command(session) == "some_value"
        assert len(mock.call_args_list) == 1

        mock.reset_mock()

        mock.return_value = "other_value"
        assert FFMPEGMuxer.command(session) == "some_value"
        assert len(mock.call_args_list) == 0

    @pytest.mark.parametrize(
        ("command", "which", "expected"),
        [
            pytest.param(None, {"ffmpeg": None}, None, id="resolver-negative"),
            pytest.param(None, {"ffmpeg": "ffmpeg"}, "ffmpeg", id="resolver-posix"),
            pytest.param(None, {"ffmpeg": "ffmpeg.exe"}, "ffmpeg.exe", id="resolver-windows"),
            pytest.param("custom", {"ffmpeg": "ffmpeg"}, None, id="custom-negative"),
            pytest.param("custom", {"ffmpeg": "ffmpeg", "custom": "custom"}, "custom", id="custom-positive"),
        ],
    )
    def test_no_cache(
        self,
        monkeypatch: pytest.MonkeyPatch,
        session: Streamlink,
        command: str | None,
        which: dict,
        expected: str | None,
    ):
        monkeypatch.setattr("streamlink.stream.ffmpegmux.which", Mock(side_effect=which.get))
        session.options.update({"ffmpeg-ffmpeg": command})
        assert FFMPEGMuxer.command(session) == expected

    @pytest.mark.parametrize(
        ("resolved", "expected"),
        [
            pytest.param(None, False, id="negative"),
            pytest.param("ffmpeg", True, id="positive"),
        ],
    )
    def test_is_usable(self, monkeypatch: pytest.MonkeyPatch, session: Streamlink, resolved: str | None, expected: bool):
        monkeypatch.setattr("streamlink.stream.ffmpegmux.which", Mock(return_value=resolved))
        assert FFMPEGMuxer.is_usable(session) is expected

    def test_log(self, monkeypatch: pytest.MonkeyPatch, caplog: pytest.LogCaptureFixture, session: Streamlink):
        monkeypatch.setattr("streamlink.stream.ffmpegmux.which", Mock(return_value=None))

        assert not FFMPEGMuxer.is_usable(session)
        assert [(record.module, record.levelname, record.message) for record in caplog.records] == [
            (
                "ffmpegmux",
                "warning",
                "No valid FFmpeg binary was found. See the --ffmpeg-ffmpeg option.",
            ),
            (
                "ffmpegmux",
                "warning",
                "Muxing streams is unsupported! Only a subset of the available streams can be returned!",
            ),
        ]

        caplog.records.clear()
        assert not FFMPEGMuxer.is_usable(session)
        assert len(caplog.records) == 0

    def test_no_log(self, monkeypatch: pytest.MonkeyPatch, caplog: pytest.LogCaptureFixture, session: Streamlink):
        monkeypatch.setattr("streamlink.stream.ffmpegmux.which", Mock(return_value="foo"))
        assert FFMPEGMuxer.is_usable(session)
        assert len(caplog.records) == 0

    def test_validate_success(self, monkeypatch: pytest.MonkeyPatch, caplog: pytest.LogCaptureFixture, session: Streamlink):
        session.options.update({"ffmpeg-no-validation": False})

        class MyFFmpegVersionOutput(FFmpegVersionOutput):
            def run(self):
                self.onstdout(0, "ffmpeg version 0.0.0 suffix")
                self.onstdout(1, "foo")
                self.onstdout(2, "bar")
                return True

        mock_versionoutput = Mock(side_effect=MyFFmpegVersionOutput)
        monkeypatch.setattr("streamlink.stream.ffmpegmux.which", Mock(return_value="/usr/bin/ffmpeg"))
        monkeypatch.setattr("streamlink.stream.ffmpegmux.FFmpegVersionOutput", mock_versionoutput)

        result = FFMPEGMuxer.command(session)
        assert result == "/usr/bin/ffmpeg"
        assert mock_versionoutput.call_args_list == [
            call(["/usr/bin/ffmpeg", "-version"], timeout=4.0),
        ]
        assert [(record.module, record.levelname, record.message) for record in caplog.records] == [
            ("ffmpegmux", "trace", "Querying FFmpeg version: ['/usr/bin/ffmpeg', '-version']"),
            ("ffmpegmux", "debug", "ffmpeg version 0.0.0 suffix"),
            ("ffmpegmux", "debug", " foo"),
            ("ffmpegmux", "debug", " bar"),
        ]

    @pytest.mark.parametrize(
        ("timeout_value", "expected_timeout"),
        [
            pytest.param(None, 4.0, id="default"),
            pytest.param(9.5, 9.5, id="custom"),
        ],
    )
    def test_validate_timeout(self, monkeypatch: pytest.MonkeyPatch, session: Streamlink, timeout_value, expected_timeout):
        session.options.update({
            "ffmpeg-no-validation": False,
            "ffmpeg-validation-timeout": timeout_value,
        })

        class MyFFmpegVersionOutput(FFmpegVersionOutput):
            def run(self):
                self.onstdout(0, "ffmpeg version 0.0.0 custom")
                return True

        mock_versionoutput = Mock(side_effect=MyFFmpegVersionOutput)
        monkeypatch.setattr("streamlink.stream.ffmpegmux.which", Mock(return_value="/usr/bin/ffmpeg"))
        monkeypatch.setattr("streamlink.stream.ffmpegmux.FFmpegVersionOutput", mock_versionoutput)

        result = FFMPEGMuxer.command(session)
        assert result == "/usr/bin/ffmpeg"
        assert mock_versionoutput.call_args_list == [
            call(["/usr/bin/ffmpeg", "-version"], timeout=expected_timeout),
        ]

    def test_validate_failure(self, monkeypatch: pytest.MonkeyPatch, caplog: pytest.LogCaptureFixture, session: Streamlink):
        session.options.update({"ffmpeg-no-validation": False})

        class MyFFmpegVersionOutput(FFmpegVersionOutput):
            def run(self):
                return False

        mock_versionoutput = Mock(side_effect=MyFFmpegVersionOutput)
        monkeypatch.setattr("streamlink.stream.ffmpegmux.which", Mock(return_value="/usr/bin/ffmpeg"))
        monkeypatch.setattr("streamlink.stream.ffmpegmux.FFmpegVersionOutput", mock_versionoutput)

        result = FFMPEGMuxer.command(session)
        assert result is None
        assert mock_versionoutput.call_args_list == [
            call(["/usr/bin/ffmpeg", "-version"], timeout=4.0),
        ]
        assert [(record.module, record.levelname, record.message) for record in caplog.records] == [
            ("ffmpegmux", "trace", "Querying FFmpeg version: ['/usr/bin/ffmpeg', '-version']"),
            ("ffmpegmux", "error", "Could not validate FFmpeg!"),
            ("ffmpegmux", "error", "Unexpected FFmpeg version output while running ['/usr/bin/ffmpeg', '-version']"),
            ("ffmpegmux", "warning", "No valid FFmpeg binary was found. See the --ffmpeg-ffmpeg option."),
            ("ffmpegmux", "warning", "Muxing streams is unsupported! Only a subset of the available streams can be returned!"),
        ]


class TestFFmpegVersionOutput:
    @pytest.fixture()
    def output(self):
        output = FFmpegVersionOutput(["/usr/bin/ffmpeg", "-version"], timeout=1.0)
        assert output.command == ["/usr/bin/ffmpeg", "-version"]
        assert output.timeout == 1.0  # noqa: RUF069
        assert output.output == []
        assert output.version is None

        return output

    def test_success(self, output: FFmpegVersionOutput):
        output.onstdout(0, "ffmpeg version 0.0.0 suffix")
        assert output.output == ["ffmpeg version 0.0.0 suffix"]
        assert output.version == "0.0.0"

        output.onstdout(1, "foo")
        output.onstdout(2, "bar")
        assert output.output == ["ffmpeg version 0.0.0 suffix", "foo", "bar"]
        assert output.version == "0.0.0"

        assert output.onexit(0)

    def test_failure_stdout(self, output: FFmpegVersionOutput):
        output.onstdout(0, "invalid")
        assert output.output == []
        assert output.version is None
        assert not output.onexit(0)

    def test_failure_exitcode(self, output: FFmpegVersionOutput):
        output.onstdout(0, "ffmpeg version 0.0.0 suffix")
        assert output.output == ["ffmpeg version 0.0.0 suffix"]
        assert output.version == "0.0.0"
        assert not output.onexit(1)


class TestOpen:
    FFMPEG_ARGS_DEFAULT_BASE = ["-y", "-nostats"]
    FFMPEG_ARGS_DEFAULT_LOGLEVEL = ["-loglevel", "info"]
    FFMPEG_ARGS_DEFAULT_CODECS = ["-c:v", FFMPEGMuxer.DEFAULT_VIDEO_CODEC, "-c:a", FFMPEGMuxer.DEFAULT_AUDIO_CODEC]
    FFMPEG_ARGS_DEFAULT_FORMAT = ["-f", FFMPEGMuxer.DEFAULT_OUTPUT_FORMAT]
    FFMPEG_ARGS_DEFAULT_OUTPUT = ["pipe:1"]

    @pytest.fixture(autouse=True)
    def which(self, monkeypatch: pytest.MonkeyPatch):
        mock = Mock(return_value="ffmpeg")
        monkeypatch.setattr("streamlink.stream.ffmpegmux.which", mock)

        return mock

    @pytest.fixture()
    def popen(self, monkeypatch: pytest.MonkeyPatch):
        mock = Mock()
        monkeypatch.setattr("subprocess.Popen", mock)

        return mock

    @pytest.mark.parametrize(
        ("options", "muxer_args", "expected"),
        [
            pytest.param(
                {},
                {},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="default",
            ),
            pytest.param(
                {},
                {"loglevel": "verbose"},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    "-loglevel",
                    "verbose",
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="loglevel",
            ),
            pytest.param(
                {"ffmpeg-loglevel": "error"},
                {"loglevel": "verbose"},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    "-loglevel",
                    "error",
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="loglevel-user-override",
            ),
            pytest.param(
                {},
                {"format": "mpegts"},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    "-f",
                    "mpegts",
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="format",
            ),
            pytest.param(
                {"ffmpeg-fout": "avi"},
                {"format": "mpegts"},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    "-f",
                    "avi",
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="format-user-override",
            ),
            pytest.param(
                {},
                {"copyts": True},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    "-copyts",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="copyts",
            ),
            pytest.param(
                {"ffmpeg-copyts": True},
                {"copyts": False},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    "-copyts",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="copyts-user-override",
            ),
            pytest.param(
                {"ffmpeg-start-at-zero": False},
                {"copyts": True},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    "-copyts",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="copyts-disable-session-start-at-zero",
            ),
            pytest.param(
                {"ffmpeg-copyts": True, "ffmpeg-start-at-zero": False},
                {"copyts": False},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    "-copyts",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="copyts-disable-session-start-at-zero-user-override",
            ),
            pytest.param(
                {"ffmpeg-start-at-zero": True},
                {"copyts": True},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    "-copyts",
                    "-start_at_zero",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="copyts-enable-session-start-at-zero",
            ),
            pytest.param(
                {"ffmpeg-copyts": True, "ffmpeg-start-at-zero": True},
                {"copyts": False},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    "-copyts",
                    "-start_at_zero",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="copyts-enable-session-start-at-zero-user-override",
            ),
            pytest.param(
                {},
                {"copyts": True, "start_at_zero": False},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    "-copyts",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="copyts-disable-start-at-zero",
            ),
            pytest.param(
                {"ffmpeg-copyts": True},
                {"copyts": False, "start_at_zero": False},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    "-copyts",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="copyts-disable-start-at-zero-user-override",
            ),
            pytest.param(
                {},
                {"copyts": True, "start_at_zero": True},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    "-copyts",
                    "-start_at_zero",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="copyts-enable-start-at-zero",
            ),
            pytest.param(
                {"ffmpeg-copyts": True},
                {"copyts": False, "start_at_zero": True},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    "-copyts",
                    "-start_at_zero",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="copyts-enable-start-at-zero-user-override",
            ),
            pytest.param(
                {},
                {"vcodec": "avc"},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    "-c:v",
                    "avc",
                    "-c:a",
                    FFMPEGMuxer.DEFAULT_AUDIO_CODEC,
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="vcodec",
            ),
            pytest.param(
                {"ffmpeg-video-transcode": "divx"},
                {"vcodec": "avc"},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    "-c:v",
                    "divx",
                    "-c:a",
                    FFMPEGMuxer.DEFAULT_AUDIO_CODEC,
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="vcodec-user-override",
            ),
            pytest.param(
                {},
                {"acodec": "mp3"},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    "-c:v",
                    FFMPEGMuxer.DEFAULT_VIDEO_CODEC,
                    "-c:a",
                    "mp3",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="acodec",
            ),
            pytest.param(
                {"ffmpeg-audio-transcode": "ogg"},
                {"acodec": "mp3"},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    "-c:v",
                    FFMPEGMuxer.DEFAULT_VIDEO_CODEC,
                    "-c:a",
                    "ogg",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="acodec-user-override",
            ),
            pytest.param(
                {},
                {"vcodec": "avc", "acodec": "mp3"},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    "-c:v",
                    "avc",
                    "-c:a",
                    "mp3",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="vcodec-acodec",
            ),
            pytest.param(
                {"ffmpeg-video-transcode": "divx", "ffmpeg-audio-transcode": "ogg"},
                {"vcodec": "avc", "acodec": "mp3"},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    "-c:v",
                    "divx",
                    "-c:a",
                    "ogg",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="vcodec-acodec-user-override",
            ),
            pytest.param(
                {},
                {"maps": ["test", "test2"]},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    "-map",
                    "test",
                    "-map",
                    "test2",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="maps",
            ),
            pytest.param(
                {},
                {"metadata": {"s:a:0": ["language=eng"]}},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    "-metadata:s:a:0",
                    "language=eng",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="metadata-stream-audio",
            ),
            pytest.param(
                {},
                {"metadata": {None: ["title=test"]}},
                [
                    *FFMPEG_ARGS_DEFAULT_BASE,
                    *FFMPEG_ARGS_DEFAULT_LOGLEVEL,
                    *FFMPEG_ARGS_DEFAULT_CODECS,
                    "-metadata",
                    "title=test",
                    *FFMPEG_ARGS_DEFAULT_FORMAT,
                    *FFMPEG_ARGS_DEFAULT_OUTPUT,
                ],
                id="metadata-title",
            ),
        ],
    )
    def test_ffmpeg_args(
        self,
        session: Streamlink,
        popen: Mock,
        options: dict,
        muxer_args: dict,
        expected: list,
    ):
        session.options.update(options)
        streamio = FFMPEGMuxer(session, **muxer_args)

        streamio.open()
        assert popen.call_args_list == [
            call(
                ["ffmpeg", *expected],
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.DEVNULL,
            ),
        ]

        streamio.close()

    def test_stderr(self, monkeypatch: pytest.MonkeyPatch, session: Streamlink, popen: Mock):
        mock_stderr = Mock()
        session.options.update({"ffmpeg-verbose": True})
        monkeypatch.setattr("streamlink.stream.ffmpegmux.sys.stderr", mock_stderr)

        streamio = FFMPEGMuxer(session)
        streamio.open()
        assert popen.call_args_list[0][1]["stderr"] is mock_stderr

        streamio.close()
        assert mock_stderr.close.call_count == 0

    @pytest.mark.parametrize(
        ("options", "side_effect"),
        [
            pytest.param({"ffmpeg-verbose-path": "foo"}, None, id="verbose-path"),
            pytest.param({"ffmpeg-verbose-path": "foo", "ffmpeg-verbose": True}, None, id="verbose-path priority"),
            pytest.param({"ffmpeg-verbose-path": "foo"}, OSError, id="OSError on close"),
        ],
    )
    def test_stderr_path(
        self,
        monkeypatch: pytest.MonkeyPatch,
        session: Streamlink,
        popen: Mock,
        options: dict,
        side_effect: type[Exception] | None,
    ):
        mock_path = Mock()
        session.options.update(options)
        monkeypatch.setattr("streamlink.stream.ffmpegmux.Path", mock_path)

        file: Mock = mock_path("foo").expanduser().open("w")
        file.close.side_effect = side_effect
        streamio = FFMPEGMuxer(session)

        streamio.open()
        assert popen.call_args_list[0][1]["stderr"] is file
        assert file.close.call_count == 0

        streamio.close()
        assert file.close.call_count == 1



from __future__ import annotations

from contextlib import AbstractContextManager, nullcontext
from pathlib import Path
from signal import SIGTERM

import pytest
import trio

from streamlink.compat import BaseExceptionGroup, is_win32
from streamlink.webbrowser.exceptions import WebbrowserError
from streamlink.webbrowser.webbrowser import Webbrowser


class _FakeWebbrowser(Webbrowser):
    @classmethod
    def launch_args(cls) -> list[str]:
        return ["foo", "bar"]


class TestInit:
    @pytest.mark.parametrize(
        ("executable", "resolve_executable", "raises"),
        [
            pytest.param(
                None,
                None,
                pytest.raises(WebbrowserError, match=r"^Could not find web browser executable: Please set the path "),
                id="Failure with unset path",
            ),
            pytest.param(
                "custom",
                None,
                pytest.raises(WebbrowserError, match=r"^Invalid web browser executable: custom$"),
                id="Failure with custom path",
            ),
            pytest.param(
                None,
                "default",
                nullcontext(),
                id="Success with default path",
            ),
            pytest.param(
                "custom",
                "custom",
                nullcontext(),
                id="Success with custom path",
            ),
        ],
        indirect=["resolve_executable"],
    )
    def test_resolve_executable(self, resolve_executable, executable: str | None, raises: nullcontext):
        with raises:
            Webbrowser(executable=executable)

    def test_arguments(self):
        webbrowser = _FakeWebbrowser()
        assert webbrowser.executable == "default"
        assert webbrowser.arguments == ["foo", "bar"]
        assert webbrowser.arguments is not _FakeWebbrowser.launch_args()


class TestLaunch:
    @pytest.mark.trio()
    async def test_terminate_on_nursery_exit(self, caplog: pytest.LogCaptureFixture, webbrowser_launch):
        process: trio.Process
        async with webbrowser_launch() as (_nursery, process):
            assert process.poll() is None, "process is still running"

        assert process.poll() == (1 if is_win32 else -SIGTERM), "Process has been terminated"
        assert [(record.name, record.levelname, record.msg) for record in caplog.records] == [
            ("streamlink.webbrowser.webbrowser", "debug", "Waiting for web browser process to terminate"),
        ]

    @pytest.mark.trio()
    async def test_terminate_on_nursery_cancellation(self, caplog: pytest.LogCaptureFixture, webbrowser_launch):
        nursery: trio.Nursery
        process: trio.Process
        async with webbrowser_launch() as (nursery, process):
            assert process.poll() is None, "process is still running"
            nursery.cancel_scope.cancel()

        assert process.poll() == (1 if is_win32 else -SIGTERM), "Process has been terminated"
        assert [(record.name, record.levelname, record.msg) for record in caplog.records] == [
            ("streamlink.webbrowser.webbrowser", "debug", "Waiting for web browser process to terminate"),
        ]

    @pytest.mark.trio()
    async def test_terminate_on_nursery_timeout(self, caplog: pytest.LogCaptureFixture, mock_clock, webbrowser_launch):
        process: trio.Process
        async with webbrowser_launch(timeout=10) as (_nursery, process):
            assert process.poll() is None, "process is still running"
            mock_clock.jump(20)
            await trio.lowlevel.checkpoint()

        assert process.poll() == (1 if is_win32 else -SIGTERM), "Process has been terminated"
        assert [(record.name, record.levelname, record.msg) for record in caplog.records] == [
            ("streamlink.webbrowser.webbrowser", "warning", "Web browser task group has timed out"),
            ("streamlink.webbrowser.webbrowser", "debug", "Waiting for web browser process to terminate"),
        ]

    @pytest.mark.trio()
    @pytest.mark.parametrize("exception", [KeyboardInterrupt, SystemExit])
    async def test_propagate_keyboardinterrupt_systemexit(self, caplog: pytest.LogCaptureFixture, webbrowser_launch, exception):
        process: trio.Process
        with pytest.raises(exception) as excinfo:  # noqa: PT012
            async with webbrowser_launch() as (_nursery, process):
                assert process.poll() is None, "process is still running"
                async with trio.open_nursery():
                    raise exception()

        assert isinstance(excinfo.value.__context__, BaseExceptionGroup)
        assert process.poll() == (1 if is_win32 else -SIGTERM), "Process has been terminated"
        assert [(record.name, record.levelname, record.msg) for record in caplog.records] == [
            ("streamlink.webbrowser.webbrowser", "debug", "Waiting for web browser process to terminate"),
        ]

    @pytest.mark.trio()
    async def test_terminate_on_exception(self, caplog: pytest.LogCaptureFixture, webbrowser_launch):
        process: trio.Process
        with pytest.raises(BaseExceptionGroup) as excinfo:  # noqa: PT012
            async with webbrowser_launch() as (_nursery, process):
                assert process.poll() is None, "process is still running"
                async with trio.open_nursery():
                    raise ZeroDivisionError()

        assert excinfo.group_contains(ZeroDivisionError)
        assert process.poll() == (1 if is_win32 else -SIGTERM), "Process has been terminated"
        assert [(record.name, record.levelname, record.msg) for record in caplog.records] == [
            ("streamlink.webbrowser.webbrowser", "debug", "Waiting for web browser process to terminate"),
        ]

    # don't run on Windows, because of some weird flaky subprocess early-termination issues
    @pytest.mark.posix_only()
    @pytest.mark.trio()
    # don't check for non-zero exit codes - we don't care
    @pytest.mark.parametrize("exit_code", [0, 1])
    async def test_process_ended_early(self, caplog: pytest.LogCaptureFixture, webbrowser_launch, exit_code):
        process: trio.Process
        async with webbrowser_launch(timeout=10) as (_nursery, process):
            assert process.poll() is None, "process is still running"
            assert process.stdin
            await process.stdin.send_all(str(exit_code).encode() + b"\r\n")
            await trio.sleep(5)

        assert process.poll() == exit_code, "Process has ended with the right exit code"
        assert [(record.name, record.levelname, record.msg) for record in caplog.records] == [
            ("streamlink.webbrowser.webbrowser", "warning", "Web browser process ended early"),
        ]


def test_temp_dir():
    webbrowser = Webbrowser()
    temp_dir = webbrowser._create_temp_dir()
    assert isinstance(temp_dir, AbstractContextManager)
    with temp_dir as path:
        assert Path(path).exists()
    assert not Path(path).exists()



from streamlink.plugins.mangomolo import Mangomolo
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlMangomolo(PluginCanHandleUrl):
    __plugin__ = Mangomolo

    should_match = [
        (
            "mangomoloplayer",
            "https://player.mangomolo.com/v1/live?id=MTk1&channelid=MzU1&countries=bnVsbA==&w=100%25&h=100%25"
            + "&autoplay=true&filter=none&signature=9ea6c8ed03b8de6e339d6df2b0685f25&app_id=43",
        ),
        (
            "51comkw",
            "https://51.com.kw/live",
        ),
        (
            "51comkw",
            "https://51.com.kw/live/",
        ),
        (
            "51comkw",
            "https://51.com.kw/live/355/AlKuwait",
        ),
        (
            "51comkw",
            "https://51.com.kw/live/361/KTV2",
        ),
        (
            "51comkw",
            "https://51.com.kw/live/356/News",
        ),
        (
            "51comkw",
            "https://51.com.kw/live/362/Sport",
        ),
        (
            "51comkw",
            "https://51.com.kw/live/357/Sport2",
        ),
    ]



import importlib.util

import pytest

from streamlink.exceptions import StreamlinkDeprecationWarning


@pytest.mark.parametrize(
    "attr",
    [
        "LRUCache",
        "search_dict",
        "load_module",
        "NamedPipe",
        "parse_html",
        "parse_json",
        "parse_qsd",
        "parse_xml",
        "absolute_url",
        "prepend_www",
        "update_qsd",
        "update_scheme",
        "url_concat",
        "url_equal",
    ],
)
def test_deprecated(attr: str):
    spec = importlib.util.find_spec("streamlink.utils", "streamlink")
    assert spec
    assert spec.loader
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    with pytest.warns(StreamlinkDeprecationWarning):
        item = getattr(module, attr)

    assert item



from streamlink.plugins.bigo import Bigo
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlBigo(PluginCanHandleUrl):
    __plugin__ = Bigo

    should_match = [
        "https://www.bigo.tv/SITE_ID",
    ]



from streamlink.plugins.ceskatelevize import Ceskatelevize
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlCeskatelevize(PluginCanHandleUrl):
    __plugin__ = Ceskatelevize

    should_match_groups = [
        (("channel", "https://ct24.ceskatelevize.cz/"), {"channel": "ct24"}),
        (("channel", "https://decko.ceskatelevize.cz/"), {"channel": "decko"}),
        (("sport", "https://sport.ceskatelevize.cz/"), {}),
        (("default", "https://ceskatelevize.cz/zive/ct1"), {}),
        (("default", "https://ceskatelevize.cz/zive/ct2"), {}),
        (("default", "https://ceskatelevize.cz/zive/art"), {}),
        (("default", "https://ceskatelevize.cz/zive/ch-28"), {}),
    ]

    should_not_match = [
        "https://ceskatelevize.cz/",
        "https://www.ceskatelevize.cz/zive/",
    ]
